#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Cost functions for cost-aware acquisition functions, e.g. multi-fidelity KG.
To be used in a context where there is an objective/cost tradeoff.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable

import torch
from botorch.acquisition.objective import (
    GenericMCObjective,
    IdentityMCObjective,
    MCAcquisitionObjective,
)
from botorch.models.deterministic import DeterministicModel
from botorch.models.gpytorch import GPyTorchModel
from botorch.sampling.base import MCSampler
from pyre_extensions import none_throws
from torch import Tensor
from torch.nn import Module


class CostAwareUtility(Module, ABC):
    """Abstract base class for cost-aware utilities."""

    @abstractmethod
    def forward(
        self, X: Tensor, deltas: Tensor, sampler: MCSampler | None = None
    ) -> Tensor:
        r"""Evaluate the cost-aware utility on the candidates and improvements.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of with `q` `d`-dim design
                points each for each t-batch.
            deltas: A `num_fantasies x batch_shape`-dim Tensor of `num_fantasy`
                samples from the marginal improvement in utility over the
                current state at `X` for each t-batch.
            sampler: A sampler used for sampling from the posterior of the cost
                model. Some subclasses ignore this argument.

        Returns:
            A `num_fantasies x batch_shape`-dim Tensor of cost-transformed utilities.
        """


class GenericCostAwareUtility(CostAwareUtility):
    r"""Generic cost-aware utility wrapping a callable."""

    def __init__(self, cost: Callable[[Tensor, Tensor], Tensor]) -> None:
        r"""Generic cost-aware utility wrapping a callable.

        Args:
            cost: A callable mapping a `batch_shape x q x d'`-dim candidate set
                to a `batch_shape`-dim tensor of costs
        """
        super().__init__()
        self._cost_callable: Callable[[Tensor, Tensor], Tensor] = cost

    def forward(
        self, X: Tensor, deltas: Tensor, sampler: MCSampler | None = None
    ) -> Tensor:
        r"""Evaluate the cost function on the candidates and improvements.

        Args:
            X: A `batch_shape x q x d'`-dim Tensor of with `q` `d`-dim design
                points for each t-batch.
            deltas: A `num_fantasies x batch_shape`-dim Tensor of `num_fantasy`
                samples from the marginal improvement in utility over the
                current state at `X` for each t-batch.
            sampler: Ignored.

        Returns:
            A `num_fantasies x batch_shape`-dim Tensor of cost-weighted utilities.
        """
        return self._cost_callable(X, deltas)


class InverseCostWeightedUtility(CostAwareUtility):
    r"""A cost-aware utility using inverse cost weighting based on a model.

    Computes the cost-aware utility by inverse-weighting samples
    `U = (u_1, ..., u_N)` of the increase in utility. If `use_mean=True`, this
    uses the posterior mean `mean_cost` of the cost model, i.e.
    `weighted utility = mean(U) / mean_cost`. If `use_mean=False`, it uses
    samples `C = (c_1, ..., c_N)` from the posterior of the cost model and
    performs the inverse weighting on the sample level:
    `weighted utility = mean(u_1 / c_1, ..., u_N / c_N)`.

    Where values in (u_1, ..., u_N) are negative, or for mean(U) < 0, the
    weighted utility is instead calculated via scaling by the cost, i.e. if
    `use_mean=True`: `weighted_utility = mean(U) * mean_cost` and if
    `use_mean=False`:
    `weighted utility = mean(u_1 * c_1, u_2 / c_2, u_3 * c_3, ..., u_N / c_N)`,
    depending on whether (`u_*` >= 0), as with `u_2` and `u_N` in this case, or
    (`u_*` < 0) as with `u_1` and `u_3`.

    The cost is additive across multiple elements of a q-batch.
    """

    def __init__(
        self,
        cost_model: DeterministicModel | GPyTorchModel,
        use_mean: bool = True,
        cost_objective: MCAcquisitionObjective | None = None,
        log: bool = False,
    ) -> None:
        r"""Cost-aware utility that weights increase in utility by inverse cost.
        For negative increases in utility, the utility is instead scaled by the
        cost. See the class description for more information.

        Args:
            cost_model: A model of the cost of evaluating a candidate
                set `X`, where `X` are the same features as in the model for the
                acquisition function this is to be used with. If no cost_objective
                is specified, the outputs are required to be non-negative.
            use_mean: If True, use the posterior mean, otherwise use posterior
                samples from the cost model.
            cost_objective: If specified, transform the posterior mean / the
                posterior samples from the cost model. This can be used e.g. to
                un-transform predictions/samples of a cost model fit on the
                log-transformed cost (often done to ensure non-negativity). If the
                cost model is multi-output, then by default this will sum the cost
                across outputs. NOTE: ``cost_objective`` must output
                strictly positive values; forward will raise a ``ValueError`` otherwise.
            min_cost: A value used to clamp the cost samples so that they are not
                too close to zero, which may cause numerical issues.
        Returns:
            The inverse-cost-weighted utility.
        """
        super().__init__()
        if cost_objective is None:
            if cost_model.num_outputs == 1:
                cost_objective = IdentityMCObjective()
            else:
                # sum over outputs
                cost_objective = GenericMCObjective(lambda Y, X: Y.sum(dim=-1))

        self.cost_model = cost_model
        self.cost_objective: MCAcquisitionObjective = cost_objective
        self._use_mean = use_mean
        self._log = log

    def forward(
        self,
        X: Tensor,
        deltas: Tensor,
        sampler: MCSampler | None = None,
        X_evaluation_mask: Tensor | None = None,
    ) -> Tensor:
        r"""Evaluate the cost function on the candidates and improvements. Note
        that negative values of `deltas` are instead scaled by the cost, and not
        inverse-weighted. See the class description for more information.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of with `q` `d`-dim design
                points each for each t-batch.
            deltas: A `num_fantasies x batch_shape`-dim Tensor of `num_fantasy`
                samples from the marginal improvement in utility over the
                current state at `X` for each t-batch.
            sampler: A sampler used for sampling from the posterior of the cost
                model (required if `use_mean=False`, ignored if `use_mean=True`).
            X_evaluation_mask: A `q x m`-dim boolean tensor indicating which
                outcomes should be evaluated for each design in the batch.

        Returns:
            A `num_fantasies x batch_shape`-dim Tensor of cost-weighted utilities.
        """
        if not self._use_mean and sampler is None:
            raise RuntimeError("Must provide `sampler` if `use_mean=False`")
        if X_evaluation_mask is not None:
            # TODO: support different evaluation masks for each X. This requires
            # either passing evaluation_mask to `cost_model.posterior`
            # or assuming that evaluating `cost_model.posterior(X)` on all
            # `q` points and then only selecting the costs for relevant points
            # does not change the cost function for each point. This would not be
            # true for instance if the incremental cost of evaluating an additional
            # point decreased as the number of points increased.
            if not all(
                torch.equal(X_evaluation_mask[0], X_evaluation_mask[i])
                for i in range(1, X_evaluation_mask.shape[0])
            ):
                raise NotImplementedError(
                    "Currently, all candidates must be evaluated on the same outputs."
                )
            output_indices = X_evaluation_mask[0].nonzero().view(-1).tolist()
        else:
            output_indices = None
        cost_posterior = self.cost_model.posterior(X, output_indices=output_indices)
        if self._use_mean:
            cost = cost_posterior.mean  # batch_shape x q x m'
        else:
            # This will be of shape num_fantasies x batch_shape x q x m'
            cost = none_throws(sampler)(cost_posterior)
        cost = self.cost_objective(cost)

        # Ensure that costs are positive
        if not torch.all(cost > 0.0):
            raise ValueError(
                "Costs must be strictly positive. Consider clamping cost_objective."
            )

        # sum costs along q-batch
        cost = cost.sum(dim=-1)

        # compute and return the ratio on the sample level - If `use_mean=True`
        # this operation involves broadcasting the cost across fantasies.
        if self._log:
            # if _log is True then input deltas are in log space
            # so original deltas cannot be <= 0
            return deltas - torch.log(cost)
        else:
            # We multiply by the cost if the deltas are <= 0, see discussion #2914
            return torch.where(deltas > 0, deltas / cost, deltas * cost)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Assorted helper methods and objects for working with BoTorch models."""

from __future__ import annotations

import warnings
from collections.abc import Iterator
from contextlib import contextmanager, ExitStack

import torch
from botorch import settings
from botorch.exceptions import InputDataError, InputDataWarning
from botorch.settings import _Flag
from gpytorch import settings as gpt_settings
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from gpytorch.module import Module
from torch import Tensor


def _make_X_full(X: Tensor, output_indices: list[int], tf: int) -> Tensor:
    r"""Helper to construct input tensor with task indices.

    Args:
        X: The raw input tensor (without task information).
        output_indices: The output indices to generate (passed in via `posterior`).
        tf: The task feature index.

    Returns:
        Tensor: The full input tensor for the multi-task model, including task
            indices.
    """
    index_shape = X.shape[:-1] + torch.Size([1])
    indexers = (
        torch.full(index_shape, fill_value=i, device=X.device, dtype=X.dtype)
        for i in output_indices
    )
    X_l, X_r = X[..., :tf], X[..., tf:]
    return torch.cat(
        [torch.cat([X_l, indexer, X_r], dim=-1) for indexer in indexers], dim=-2
    )


def multioutput_to_batch_mode_transform(
    train_X: Tensor,
    train_Y: Tensor,
    num_outputs: int,
    train_Yvar: Tensor | None = None,
) -> tuple[Tensor, Tensor, Tensor | None]:
    r"""Transforms training inputs for a multi-output model.

    Used for multi-output models that internally are represented by a
    batched single output model, where each output is modeled as an
    independent batch.

    Args:
        train_X: A `n x d` or `input_batch_shape x n x d` (batch mode) tensor of
            training features.
        train_Y: A `n x m` or `target_batch_shape x n x m` (batch mode) tensor of
            training observations.
        num_outputs: number of outputs
        train_Yvar: A `n x m` or `target_batch_shape x n x m` tensor of observed
            measurement noise.

    Returns:
        3-element tuple containing

        - A `input_batch_shape x m x n x d` tensor of training features.
        - A `target_batch_shape x m x n` tensor of training observations.
        - A `target_batch_shape x m x n` tensor observed measurement noise.
    """
    # make train_Y `batch_shape x m x n`
    train_Y = train_Y.transpose(-1, -2)
    # expand train_X to `batch_shape x m x n x d`
    train_X = train_X.unsqueeze(-3).expand(
        train_X.shape[:-2] + torch.Size([num_outputs]) + train_X.shape[-2:]
    )
    if train_Yvar is not None:
        # make train_Yvar `batch_shape x m x n`
        train_Yvar = train_Yvar.transpose(-1, -2)
    return train_X, train_Y, train_Yvar


def add_output_dim(X: Tensor, original_batch_shape: torch.Size) -> tuple[Tensor, int]:
    r"""Insert the output dimension at the correct location.

    The trailing batch dimensions of X must match the original batch dimensions
    of the training inputs, but can also include extra batch dimensions.

    Args:
        X: A `(new_batch_shape) x (original_batch_shape) x n x d` tensor of
            features.
        original_batch_shape: the batch shape of the model's training inputs.

    Returns:
        2-element tuple containing

        - A `(new_batch_shape) x (original_batch_shape) x m x n x d` tensor of
            features.
        - The index corresponding to the output dimension.
    """
    X_batch_shape = X.shape[:-2]
    if len(X_batch_shape) > 0 and len(original_batch_shape) > 0:
        # check that X_batch_shape supports broadcasting or augments
        # original_batch_shape with extra batch dims
        try:
            torch.broadcast_shapes(X_batch_shape, original_batch_shape)
        except RuntimeError:
            raise RuntimeError(
                "The trailing batch dimensions of X must match the trailing "
                f"batch dimensions of the training inputs. Got {X.shape=} "
                f"and {original_batch_shape=}."
            )
    # insert `m` dimension
    X = X.unsqueeze(-3)
    output_dim_idx = max(len(original_batch_shape), len(X_batch_shape))
    return X, output_dim_idx


def check_no_nans(Z: Tensor) -> None:
    r"""Check that tensor does not contain NaN values.

    Raises an InputDataError if `Z` contains NaN values.

    Args:
        Z: The input tensor.
    """
    if torch.any(torch.isnan(Z)).item():
        raise InputDataError("Input data contains NaN values.")


def check_min_max_scaling(
    X: Tensor,
    strict: bool = False,
    atol: float = 1e-2,
    raise_on_fail: bool = False,
    ignore_dims: list[int] | None = None,
) -> None:
    r"""Check that tensor is normalized to the unit cube.

    Args:
        X: A `batch_shape x n x d` input tensor. Typically the training inputs
            of a model.
        strict: If True, require `X` to be scaled to the unit cube (rather than
            just to be contained within the unit cube).
        atol: The tolerance for the boundary check. Only used if `strict=True`.
        raise_on_fail: If True, raise an exception instead of a warning.
        ignore_dims: Subset of dimensions where the min-max scaling check is omitted.
    """
    ignore_dims = ignore_dims or []
    check_dims = list(set(range(X.shape[-1])) - set(ignore_dims))
    if len(check_dims) == 0:
        return None

    with torch.no_grad():
        X_check = X[..., check_dims]
        Xmin = torch.min(X_check, dim=-1).values
        Xmax = torch.max(X_check, dim=-1).values
        msg = None
        if strict and max(torch.abs(Xmin).max(), torch.abs(Xmax - 1).max()) > atol:
            msg = "scaled"
        if torch.any(Xmin < -atol) or torch.any(Xmax > 1 + atol):
            msg = "contained"
        if msg is not None:
            # NOTE: If you update this message, update the warning filters as well.
            # See https://github.com/meta-pytorch/botorch/pull/2508.
            msg = (
                f"Data (input features) is not {msg} to the unit cube. "
                "Please consider min-max scaling the input data."
            )
            if raise_on_fail:
                raise InputDataError(msg)
            warnings.warn(msg, InputDataWarning, stacklevel=2)


def check_standardization(
    Y: Tensor,
    atol_mean: float = 1e-2,
    atol_std: float = 1e-2,
    raise_on_fail: bool = False,
) -> None:
    r"""Check that tensor is standardized (zero mean, unit variance).

    Args:
        Y: The input tensor of shape `batch_shape x n x m`. Typically the
            train targets of a model. Standardization is checked across the
            `n`-dimension.
        atol_mean: The tolerance for the mean check.
        atol_std: The tolerance for the std check.
        raise_on_fail: If True, raise an exception instead of a warning.
    """
    with torch.no_grad():
        Ymean = torch.mean(Y, dim=-2)
        mean_not_zero = torch.abs(Ymean).max() > atol_mean
        if Y.shape[-2] <= 1:
            if mean_not_zero:
                # NOTE: If you update this message, update the warning filters as well.
                # See https://github.com/meta-pytorch/botorch/pull/2508.
                msg = (
                    f"Data (outcome observations) is not standardized (mean = {Ymean})."
                    " Please consider scaling the input to zero mean and unit variance."
                )
                if raise_on_fail:
                    raise InputDataError(msg)
                warnings.warn(msg, InputDataWarning, stacklevel=2)
        else:
            Ystd = torch.std(Y, dim=-2)
            std_not_one = torch.abs(Ystd - 1).max() > atol_std
            if mean_not_zero or std_not_one:
                # NOTE: If you update this message, update the warning filters as well.
                # See https://github.com/meta-pytorch/botorch/pull/2508.
                msg = (
                    "Data (outcome observations) is not standardized "
                    f"(std = {Ystd}, mean = {Ymean})."
                    "Please consider scaling the input to zero mean and unit variance."
                )
                if raise_on_fail:
                    raise InputDataError(msg)
                warnings.warn(msg, InputDataWarning, stacklevel=2)


def validate_input_scaling(
    train_X: Tensor,
    train_Y: Tensor,
    train_Yvar: Tensor | None = None,
    raise_on_fail: bool = False,
    ignore_X_dims: list[int] | None = None,
    check_nans_only: bool = False,
) -> None:
    r"""Helper function to validate input data to models.

    Args:
        train_X: A `n x d` or `batch_shape x n x d` (batch mode) tensor of
            training features.
        train_Y: A `n x m` or `batch_shape x n x m` (batch mode) tensor of
            training observations.
        train_Yvar: A `batch_shape x n x m` or `batch_shape x n x m` (batch mode)
            tensor of observed measurement noise.
        raise_on_fail: If True, raise an error instead of emitting a warning
            (only for normalization/standardization checks, an error is always
            raised if NaN values are present).
        ignore_X_dims: For this subset of dimensions from `{1, ..., d}`, ignore the
            min-max scaling check.
        check_nans_only: If True, only check for NaN values. Skips min-max scaling
            and standardization checks. This is used when the model is provided
            with a custom covariance module, to avoid potentially irrelevant
            warnings.

    This function is typically called inside the constructor of standard BoTorch
    models. It validates the following:
    (i) none of the inputs contain NaN values
    (ii) the training data (`train_X`) is normalized to the unit cube for all
    dimensions except those in `ignore_X_dims`.
    (iii) the training targets (`train_Y`) are standardized (zero mean, unit var)
    No checks (other than the NaN check) are performed for observed variances
    (`train_Yvar`) at this point.
    """
    if settings.validate_input_scaling.off():
        return
    check_no_nans(train_X)
    check_no_nans(train_Y)
    if train_Yvar is not None:
        check_no_nans(train_Yvar)
        if torch.any(train_Yvar < 0):
            raise InputDataError("Input data contains negative variances.")
    if not check_nans_only:
        check_min_max_scaling(
            X=train_X, raise_on_fail=raise_on_fail, ignore_dims=ignore_X_dims
        )
        check_standardization(Y=train_Y, raise_on_fail=raise_on_fail)


def mod_batch_shape(module: Module, names: list[str], b: int) -> None:
    r"""Recursive helper to modify gpytorch modules' batch shape attribute.

    Modifies the module in-place.

    Args:
        module: The module to be modified.
        names: The list of names to access the attribute. If the full name of
            the module is `"module.sub_module.leaf_module"`, this will be
            `["sub_module", "leaf_module"]`.
        b: The new size of the last element of the module's `batch_shape`
            attribute.
    """
    if len(names) == 0:
        return
    m = getattr(module, names[0])
    if len(names) == 1 and hasattr(m, "batch_shape") and len(m.batch_shape) > 0:
        m.batch_shape = m.batch_shape[:-1] + torch.Size([b] if b > 0 else [])
    else:
        mod_batch_shape(module=m, names=names[1:], b=b)


@contextmanager
def gpt_posterior_settings():
    r"""Context manager for settings used for computing model posteriors."""
    with ExitStack() as es:
        if gpt_settings.debug.is_default():
            es.enter_context(gpt_settings.debug(False))
        if gpt_settings.fast_pred_var.is_default():
            es.enter_context(gpt_settings.fast_pred_var())
        es.enter_context(
            gpt_settings.detach_test_caches(settings.propagate_grads.off())
        )
        yield


def detect_duplicates(
    X: Tensor,
    rtol: float = 0,
    atol: float = 1e-8,
) -> Iterator[tuple[int, int]]:
    """Returns an iterator over index pairs `(duplicate index, original index)` for all
    duplicate entries of `X`. Supporting 2-d Tensor only.

    Args:
        X: the datapoints tensor with potential duplicated entries
        rtol: relative tolerance
        atol: absolute tolerance
    """
    if len(X.shape) != 2:
        raise ValueError("X must have 2 dimensions.")

    tols = atol
    if rtol:
        rval = X.abs().max(dim=-1, keepdim=True).values
        tols = tols + rtol * rval.max(rval.transpose(-1, -2))

    n = X.shape[-2]
    dist = torch.full((n, n), float("inf"), device=X.device, dtype=X.dtype)
    dist[torch.triu_indices(n, n, offset=1).unbind()] = torch.nn.functional.pdist(
        X, p=float("inf")
    )
    return (
        (i, int(j))
        # pyre-fixme[19]: Expected 1 positional argument.
        for diff, j, i in zip(*(dist - tols).min(dim=-2), range(n))
        if diff < 0
    )


def consolidate_duplicates(
    X: Tensor, Y: Tensor, rtol: float = 0.0, atol: float = 1e-8
) -> tuple[Tensor, Tensor, Tensor]:
    """Drop duplicated Xs and update the indices tensor Y accordingly.
    Supporting 2d Tensor only as in batch mode block design is not guaranteed.

    Args:
        X: the datapoints tensor
        Y: the index tensor to be updated (e.g., pairwise comparisons)
        rtol: relative tolerance
        atol: absolute tolerance

    Returns:
        consolidated_X: the consolidated X
        consolidated_Y: the consolidated Y (e.g., pairwise comparisons indices)
        new_indices: new index of each original item in X, a tensor of size X.shape[-2]
    """
    if len(X.shape) != 2:
        raise ValueError("X must have 2 dimensions.")

    n = X.shape[-2]
    dup_map = dict(detect_duplicates(X=X, rtol=rtol, atol=atol))

    # Handle edge cases conservatively
    # If a item is in both dup set and kept set, do not remove it
    common_set = set(dup_map.keys()).intersection(dup_map.values())
    for k in list(dup_map.keys()):
        if k in common_set or dup_map[k] in common_set:
            del dup_map[k]

    if dup_map:
        dup_indices, kept_indices = zip(*dup_map.items())

        unique_indices = sorted(set(range(n)) - set(dup_indices))

        # After dropping the duplicates,
        # the kept ones' indices may also change by being shifted up
        new_idx_map = dict(zip(unique_indices, range(len(unique_indices))))
        new_indices_for_dup = (new_idx_map[idx] for idx in kept_indices)
        new_idx_map.update(dict(zip(dup_indices, new_indices_for_dup)))
        consolidated_X = X[list(unique_indices), :]
        consolidated_Y = torch.tensor(
            [[new_idx_map[item.item()] for item in row] for row in Y.unbind()],
            dtype=torch.long,
            device=Y.device,
        )
        new_indices = (
            torch.arange(n, dtype=torch.long)
            .apply_(lambda x: new_idx_map[x])
            .to(Y.device)
        )
        return consolidated_X, consolidated_Y, new_indices
    else:
        return X, Y, torch.arange(n, device=Y.device, dtype=Y.dtype)


class fantasize(_Flag):
    r"""A flag denoting whether we are currently in a `fantasize` context."""

    _state: bool = False


def get_task_value_remapping(
    observed_task_values: Tensor,
    all_task_values: Tensor,
    dtype: torch.dtype,
    default_task_value: int | None,
) -> Tensor | None:
    """Construct an mapping of observed task values to contiguous int-valued floats.

    Args:
        observed_task_values: A sorted long-valued tensor of task values.
        all_task_values: A sorted long-valued tensor of task values.
        dtype: The dtype of the model inputs (e.g. `X`), which the new
            task values should have mapped to (e.g. float, double).
        default_task_value: The default task value to use for missing task values.

    Returns:
        A tensor of shape `task_values.max() + 1` that maps task values
        to new task values. The indexing operation `mapper[task_value]`
        will produce a tensor of new task values, of the same shape as
        the original. The elements of the `mapper` tensor that do not
        appear in the original `task_values` are mapped to `nan`. The
        return value will be `None`, when the task values are contiguous
        integers starting from zero.
    """
    if dtype not in (torch.float, torch.double):
        raise ValueError(f"dtype must be torch.float or torch.double, but got {dtype}.")
    task_range = torch.arange(
        len(observed_task_values),
        dtype=all_task_values.dtype,
        device=all_task_values.device,
    )
    mapper = None

    if default_task_value is None:
        fill_value = float("nan")
    else:
        mask = observed_task_values == default_task_value
        if not mask.any():
            fill_value = float("nan")
        else:
            idx = mask.nonzero().item()
            fill_value = task_range[idx]
    # if not all tasks are observed or they are not contiguous integers
    # then map them to contiguous integers
    if not torch.equal(task_range, all_task_values):
        # Create a tensor that maps task values to new task values.
        # The number of tasks should be small, so this should be quite efficient.
        mapper = torch.full(
            (int(all_task_values.max().item()) + 1,),
            fill_value,
            dtype=dtype,
            device=all_task_values.device,
        )
        mapper[observed_task_values] = task_range.to(dtype=dtype)
    return mapper


def extract_targets_and_noise_single_output(model) -> tuple[Tensor, Tensor | None]:
    r"""Extract targets and noise variance for single-output models (m=1).

    Args:
        model: A GPyTorch model.

    Returns:
        A tuple of (Y, Yvar) where Y and Yvar have shape [batch_shape] x n x 1.
    """
    Y = model.train_targets.unsqueeze(-1)
    Yvar = None
    if isinstance(model.likelihood, FixedNoiseGaussianLikelihood):
        Yvar = model.likelihood.noise_covar.noise.unsqueeze(-1)
    return Y, Yvar


def restore_targets_and_noise_single_output(
    model, Y: Tensor, Yvar: Tensor | None, strict: bool
) -> None:
    r"""Restore targets and noise variance for single-output models (m=1).

    Args:
        model: A GPyTorch model.
        Y: Targets tensor in shape [batch_shape] x n x 1.
        Yvar: Optional noise variance tensor in shape [batch_shape] x n x 1.
        strict: Whether to strictly enforce shape constraints.
    """
    Y = Y.squeeze(-1)
    if Yvar is not None and isinstance(model.likelihood, FixedNoiseGaussianLikelihood):
        Yvar = Yvar.squeeze(-1)
        model.likelihood.noise_covar.noise = Yvar
    model.set_train_data(targets=Y, strict=strict)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from botorch.utils.testing import BotorchTestCase
from botorch_community.models.vblls import VBLLModel
from botorch_community.posteriors.bll_posterior import BLLPosterior
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from linear_operator.operators import to_linear_operator
from torch.distributions import Normal


def _get_vbll_model(num_inputs: int = 2, num_hidden: int = 3, **tkwargs) -> VBLLModel:
    test_backbone = torch.nn.Sequential(
        torch.nn.Linear(num_inputs, num_hidden),
        torch.nn.ReLU(),
        torch.nn.Linear(num_hidden, num_hidden),
    ).to(**tkwargs)
    model = VBLLModel(backbone=test_backbone, hidden_features=num_hidden)
    return model


class TestBLLPosterior(BotorchTestCase):
    def setUp(self):
        """Set up common test components."""
        super().setUp()
        self.tkwargs = {"dtype": torch.float64, "device": self.device}
        self.X = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]], **self.tkwargs)
        self.num_X = self.X.shape[0]
        self.output_dim = 1
        self.model = _get_vbll_model(num_inputs=2, **self.tkwargs)

        # Create a proper MultivariateNormal distribution for GPyTorchPosterior
        mean = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)

        # Create a positive-definite covariance matrix
        cov_diag = torch.tensor([0.1, 0.2, 0.3], **self.tkwargs)
        cov = torch.diag(cov_diag)

        self.distribution = MultivariateNormal(mean, cov)

        self.bll_posterior = BLLPosterior(
            distribution=self.distribution,
            model=self.model,
            X=self.X,
            output_dim=self.model.num_outputs,
        )

    def test_initialization(self):
        """Test that BLLPosterior initializes correctly."""
        self.assertEqual(self.bll_posterior.distribution, self.distribution)
        self.assertEqual(self.bll_posterior.model, self.model)
        self.assertEqual(self.bll_posterior.output_dim, self.output_dim)
        self.assertTrue(torch.equal(self.bll_posterior.X, self.X))

    def test_rsample_default(self):
        """Test rsample method with default sample_shape."""
        samples = self.bll_posterior.rsample()
        expected_shape = torch.Size([1, self.num_X, 1])
        self.assertEqual(samples.shape, expected_shape)

    def test_rsample_with_explicit_shape(self):
        """Test rsample method with custom sample_shape."""
        for sample_shape in ([3], [2, 3]):
            samples = self.bll_posterior.rsample(sample_shape=torch.Size(sample_shape))
            expected_shape = torch.Size(sample_shape + [self.num_X, 1])
            self.assertEqual(samples.shape, expected_shape)

    def test_mean(self):
        """Test mean property."""
        mean = self.bll_posterior.mean
        expected_shape = torch.Size([self.num_X, 1])
        self.assertEqual(mean.shape, expected_shape)

    def test_variance(self):
        """Test variance property."""
        variance = self.bll_posterior.variance
        expected_shape = torch.Size([self.num_X, 1])
        self.assertEqual(variance.shape, expected_shape)

    def test_device(self):
        """Test device property."""
        self.assertEqual(self.bll_posterior.device, self.device)

    def test_dtype(self):
        """Test dtype property."""
        self.assertEqual(self.bll_posterior.dtype, self.tkwargs["dtype"])

    def test_mvn_property(self):
        """Test the inherited mvn property."""
        self.assertIs(self.bll_posterior.mvn, self.distribution)

    def test_base_sample_shape(self):
        """Test the inherited base_sample_shape property."""
        expected_shape = (
            self.distribution.batch_shape + self.distribution.base_sample_shape
        )
        self.assertEqual(self.bll_posterior.base_sample_shape, expected_shape)

    def test_batch_range(self):
        """Test the inherited batch_range property."""
        self.assertEqual(self.bll_posterior.batch_range, (0, -1))

    def test_extended_shape(self):
        """Test the inherited _extended_shape method."""
        sample_shape = torch.Size([4, 2])
        extended_shape = self.bll_posterior._extended_shape(sample_shape)
        base_shape = (
            self.bll_posterior.batch_shape
            + self.bll_posterior.event_shape
            + torch.Size([self.output_dim])
        )
        expected_shape = sample_shape + base_shape
        self.assertEqual(extended_shape, expected_shape)

    def test_rsample_from_base_samples(self):
        """Test the inherited rsample_from_base_samples method."""
        sample_shapes = [torch.Size([4]), torch.Size([2, 3])]
        for sample_shape in sample_shapes:
            base_sample_shape = self.bll_posterior.base_sample_shape
            base_samples = torch.randn(sample_shape + base_sample_shape, **self.tkwargs)

            samples = self.bll_posterior.rsample_from_base_samples(
                sample_shape=sample_shape, base_samples=base_samples
            )
            expected_shape = self.bll_posterior._extended_shape(sample_shape)
            self.assertEqual(samples.shape, expected_shape)

    def test_quantile(self):
        """Test the inherited quantile method."""
        value_single = torch.tensor(0.5, **self.tkwargs)
        quantiles_single = self.bll_posterior.quantile(value_single)
        self.assertTrue(torch.allclose(quantiles_single, self.bll_posterior.mean))

        value_multi = torch.tensor([0.25, 0.5, 0.75], **self.tkwargs)
        quantiles_multi = self.bll_posterior.quantile(value_multi)
        marginal = Normal(
            loc=self.bll_posterior.mean, scale=self.bll_posterior.variance.sqrt()
        )
        expected_quantiles = marginal.icdf(
            value_multi.view(-1, 1, 1)
        )  # [3] -> [3, 1, 1]
        self.assertEqual(
            quantiles_multi.shape, torch.Size([3, self.num_X, self.output_dim])
        )
        self.assertTrue(torch.allclose(quantiles_multi, expected_quantiles))

    def test_density(self):
        """Test the inherited density method."""
        marginal = Normal(
            loc=self.bll_posterior.mean, scale=self.bll_posterior.variance.sqrt()
        )
        value_single = torch.tensor(1.0, **self.tkwargs)
        density_single = self.bll_posterior.density(value_single)
        expected_density_single = marginal.log_prob(value_single).exp()
        self.assertTrue(torch.allclose(density_single, expected_density_single))

        value_multi = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        density_multi = self.bll_posterior.density(value_multi)
        expected_density_multi = marginal.log_prob(
            value_multi.view(-1, 1, 1)
        ).exp()  # [3] -> [3, 1, 1]

        self.assertTrue(torch.allclose(density_multi, expected_density_multi))
        self.assertEqual(
            density_multi.shape, torch.Size([3, self.num_X, self.output_dim])
        )

    def test_multi_output(self):
        """Test with multiple output dimensions."""
        output_dim = 2

        # Create a model with multiple outputs
        backbone = torch.nn.Sequential(
            torch.nn.Linear(2, 3),
            torch.nn.ReLU(),
            torch.nn.Linear(3, 3),
        ).to(**self.tkwargs)

        model = VBLLModel(backbone=backbone, hidden_features=3, out_features=output_dim)

        # Create mean tensor with shape [batch_size, output_dim]
        mean = torch.rand(3, 2, **self.tkwargs)
        variance = 1 + torch.rand(3, 2, **self.tkwargs)
        covar = variance.view(-1).diag()
        distribution = MultitaskMultivariateNormal(mean, to_linear_operator(covar))
        bll_posterior = BLLPosterior(
            distribution=distribution, model=model, X=self.X, output_dim=output_dim
        )

        # Test rsample
        samples = bll_posterior.rsample()
        expected_shape = torch.Size(
            [1, self.num_X, output_dim]
        )  # [samples, batch, output_dim]
        self.assertEqual(samples.shape, expected_shape)

        # Test rsample with explicit shape
        for sample_shape in ([3], [2, 3]):
            samples = bll_posterior.rsample(sample_shape=torch.Size(sample_shape))

            # Check shape: should be [sample_shape, num_X, output_dim]
            expected_shape = torch.Size(sample_shape + [self.num_X, output_dim])
            self.assertEqual(samples.shape, expected_shape)

        # Test mean with multi-output
        multi_mean = bll_posterior.mean
        expected_mean_shape = torch.Size(
            [self.num_X, output_dim]
        )  # [batch, output_dim]
        self.assertEqual(multi_mean.shape, expected_mean_shape)

        # check mean values
        self.assertTrue(torch.allclose(multi_mean, mean))

        # Test variance with multi-output
        multi_variance = bll_posterior.variance
        expected_var_shape = torch.Size([self.num_X, output_dim])  # [batch, output_dim]
        self.assertEqual(multi_variance.shape, expected_var_shape)

        # check variance values
        expected_variance = variance
        self.assertTrue(torch.allclose(multi_variance, expected_variance))

        self.assertEqual(bll_posterior.batch_range, (0, -2))
        value = torch.tensor(0.5, **self.tkwargs)
        quantiles = bll_posterior.quantile(value)
        self.assertTrue(torch.allclose(quantiles, bll_posterior.mean))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Batch acquisition functions using the reparameterization trick in combination
with (quasi) Monte-Carlo sampling. See [Rezende2014reparam]_, [Wilson2017reparam]_ and
[Balandat2020botorch]_.

References

.. [Rezende2014reparam]
    D. J. Rezende, S. Mohamed, and D. Wierstra. Stochastic backpropagation and
    approximate inference in deep generative models. ICML 2014.

.. [Wilson2017reparam]
    J. T. Wilson, R. Moriconi, F. Hutter, and M. P. Deisenroth.
    The reparameterization trick for acquisition functions. ArXiv 2017.
"""

from __future__ import annotations

import math
from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import deepcopy
from functools import partial
from typing import Protocol

import torch
from botorch.acquisition.acquisition import AcquisitionFunction, MCSamplerMixin
from botorch.acquisition.cached_cholesky import CachedCholeskyMCSamplerMixin
from botorch.acquisition.objective import (
    ConstrainedMCObjective,
    IdentityMCObjective,
    MCAcquisitionObjective,
    PosteriorTransform,
)
from botorch.acquisition.utils import (
    compute_best_feasible_objective,
    prune_inferior_points,
    repeat_to_match_aug_dim,
)
from botorch.exceptions.errors import UnsupportedError
from botorch.exceptions.warnings import legacy_ei_numerics_warning
from botorch.models.model import Model
from botorch.sampling.base import MCSampler
from botorch.utils.objective import compute_smoothed_feasibility_indicator
from botorch.utils.transforms import (
    concatenate_pending_points,
    is_ensemble,
    match_batch_shape,
    t_batch_mode_transform,
)
from torch import Tensor


class MCAcquisitionFunction(AcquisitionFunction, MCSamplerMixin, ABC):
    r"""
    Abstract base class for Monte-Carlo based batch acquisition functions.
    """

    def __init__(
        self,
        model: Model,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
    ) -> None:
        r"""
        Args:
            model: A fitted model.
            sampler: The sampler used to draw base samples. If not given,
                a sampler is generated on the fly within the
                `get_posterior_samples` method using
                `botorch.sampling.get_sampler`.
                NOTE: For posteriors that do not support base samples,
                a sampler compatible with intended use case must be provided.
                See `ForkedRNGSampler` and `StochasticSampler` as examples.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape, m x d`-dim Tensor of `m` design points
                that have points that have been submitted for function evaluation
                but have not yet been evaluated.
        """
        super().__init__(model=model)
        MCSamplerMixin.__init__(self, sampler=sampler)
        if objective is None and model.num_outputs != 1:
            if posterior_transform is None:
                raise UnsupportedError(
                    "Must specify an objective or a posterior transform when using "
                    "a multi-output model."
                )
            elif not posterior_transform.scalarize:
                raise UnsupportedError(
                    "If using a multi-output model without an objective, "
                    "posterior_transform must scalarize the output."
                )
        if objective is None:
            objective = IdentityMCObjective()
        self.posterior_transform = posterior_transform
        self.objective: MCAcquisitionObjective = objective
        self.set_X_pending(X_pending)

    def _get_samples_and_objectives(self, X: Tensor) -> tuple[Tensor, Tensor]:
        """Computes posterior samples and objective values at input X.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of model inputs.

        Returns:
            A two-tuple `(samples, obj)`, where `samples` is a tensor of posterior
            samples with shape `sample_shape x batch_shape x q x m`, and `obj` is a
            tensor of MC objective values with shape `sample_shape x batch_shape x q`.
        """
        posterior = self.model.posterior(
            X=X, posterior_transform=self.posterior_transform
        )
        samples = self.get_posterior_samples(posterior)
        obj = self.objective(samples=samples, X=X)
        return samples, obj

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Takes in a `batch_shape x q x d` X Tensor of t-batches with `q` `d`-dim
        design points each, and returns a Tensor with shape `batch_shape'`, where
        `batch_shape'` is the broadcasted batch shape of model and input `X`. Should
        utilize the result of `set_X_pending` as needed to account for pending function
        evaluations.
        """
        pass  # pragma: no cover


class SampleReductionProtocol(Protocol):
    """For static type check of SampleReducingMCAcquisitionFunction's mc_reduction."""

    @staticmethod
    def __call__(X: Tensor, *, dim: torch.Size) -> Tensor:
        pass  # pragma: no cover


class SampleReducingMCAcquisitionFunction(MCAcquisitionFunction):
    r"""MC-based batch acquisition function that reduces across samples and implements
    a general treatment of outcome constraints.

    This class's `forward` computes the - possibly constrained - acquisition value by
    (1) computing the unconstrained utility for each MC sample using `_sample_forward`,
    (2) weighing the utility values by the constraint indicator per MC sample, and
    (3) reducing (e.g. averaging) the weighted utility values over the MC dimension.

    NOTE: Do *NOT* override the `forward` method, unless you have thought about it well.

    `forward` is implemented generically to incorporate constraints in a principled way,
    and takes care of reducing over the Monte Carlo and batch dimensions via the
    `sample_reduction` and `q_reduction` arguments, which default to `torch.mean` and
    `torch.max`, respectively.

    In order to implement a custom SampleReducingMCAcquisitionFunction, we only need to
    implement the `_sample_forward(obj: Tensor) -> Tensor` method, which maps objective
    samples to acquisition utility values without reducing the Monte Carlo and batch
    (i.e. q) dimensions (see details in the docstring of `_sample_forward`).

    A note on design choices:

    The primary purpose of `SampleReducingMCAcquisitionFunction`is to support outcome
    constraints. On the surface, designing a wrapper `ConstrainedMCAcquisitionFunction`
    could be an elegant solution to this end, but it would still require the acquisition
    functions to implement a `_sample_forward` method to weigh acquisition utilities at
    the sample level. Further, `qNoisyExpectedImprovement` is a special case that is
    hard to encompass in this pattern, since it requires the computation of the best
    *feasible* objective, which requires access to the constraint functions. However,
    if the constraints are stored in a wrapper class, they will be inaccessible to the
    forward pass. These problems are circumvented by the design of this class.
    """

    _log: bool = False  # whether the acquisition utilities are in log-space

    def __init__(
        self,
        model: Model,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        sample_reduction: SampleReductionProtocol = torch.mean,
        q_reduction: SampleReductionProtocol = torch.amax,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
        fat: list[bool | None] | bool = False,
    ):
        r"""Constructor of SampleReducingMCAcquisitionFunction.

        Args:
            model: A fitted model.
            sampler: The sampler used to draw base samples. If not given, a
                sampler is generated on the fly within the
                `get_posterior_samples` method using
                `botorch.sampling.get_sampler`.
                NOTE: For posteriors that do not support base samples,
                a sampler compatible with intended use case must be provided.
                See `ForkedRNGSampler` and `StochasticSampler` as examples.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
                NOTE: `ConstrainedMCObjective` for outcome constraints is deprecated in
                favor of passing the `constraints` directly to this constructor.
            posterior_transform: A `PosteriorTransform` (optional).
            X_pending: A `batch_shape, m x d`-dim Tensor of `m` design points
                that have points that have been submitted for function evaluation
                but have not yet been evaluated.
            sample_reduction: A callable that takes in a `sample_shape x batch_shape`
                Tensor of acquisition utility values, a keyword-argument `dim` that
                specifies the sample dimensions to reduce over, and returns a
                `batch_shape`-dim Tensor of acquisition values.
            q_reduction: A callable that takes in a `sample_shape x batch_shape x q`
                Tensor of acquisition utility values, a keyword-argument `dim` that
                specifies the q dimension to reduce over (i.e. -1), and returns a
                `sample_shape x batch_shape`-dim Tensor of acquisition values.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are considered satisfied if the output is less than zero.
                NOTE: Constraint-weighting is only compatible with non-negative
                acquistion utilities, e.g. all improvement-based acquisition functions.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. For more details, on this
                parameter, see the docs of `compute_smoothed_feasibility_indicator`.
            fat: Wether to apply a fat-tailed smooth approximation to the feasibility
                indicator or the canonical sigmoid approximation. For more details,
                on this parameter, see the docs of
                `compute_smoothed_feasibility_indicator`.
        """
        if constraints is not None and isinstance(objective, ConstrainedMCObjective):
            raise ValueError(
                "ConstrainedMCObjective as well as constraints passed to constructor."
                "Choose one or the other, preferably the latter."
            )
        # TODO: deprecate ConstrainedMCObjective
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
        )
        # Shall the need arise, sample_dim could be exposed in the constructor.
        sample_dim = tuple(range(len(self.sample_shape)))
        # the ensemble dimension is assumed to be the last dimension
        if is_ensemble(model):
            sample_dim = sample_dim + (-1,)
        self._sample_reduction = partial(sample_reduction, dim=sample_dim)
        self._q_reduction = partial(q_reduction, dim=-1)
        self._constraints = constraints
        self._eta = eta
        self._fat = fat

    @concatenate_pending_points
    @t_batch_mode_transform()
    def forward(self, X: Tensor) -> Tensor:
        r"""Computes the acquisition value associated with the input `X`. Weighs the
        acquisition utility values by smoothed constraint indicators if `constraints`
        was passed to the constructor of the class. Applies `self.sample_reduction` and
        `self.q_reduction` to reduce over the Monte Carlo and batch (q) dimensions.

        NOTE: Do *NOT* override the `forward` method for a custom acquisition function.
        Instead, implement the `_sample_forward` method. See the docstring of this class
        for details.

        Args:
            X: A `batch_shape x q x d` Tensor of t-batches with `q` `d`-dim
                design points each.

        Returns:
            A Tensor with shape `batch_shape'`, where `batch_shape'` is the broadcasted
            batch shape of model and input `X`.
        """
        non_reduced_acqval = self._non_reduced_forward(X=X)
        return self._sample_reduction(self._q_reduction(non_reduced_acqval))

    def _non_reduced_forward(self, X: Tensor) -> Tensor:
        """Compute the constrained acquisition values at the MC-sample, q level.

        Args:
            X: A `batch_shape x q x d` Tensor of t-batches with `q` `d`-dim
                design points each.

        Returns:
            A Tensor with shape `sample_sample x batch_shape x q`.
        """
        samples, obj = self._get_samples_and_objectives(X)
        samples = repeat_to_match_aug_dim(target_tensor=samples, reference_tensor=obj)
        acqval = self._sample_forward(obj)  # `sample_sample x batch_shape x q`
        return self._apply_constraints(acqval=acqval, samples=samples)

    @abstractmethod
    def _sample_forward(self, obj: Tensor) -> Tensor:
        """Evaluates the acquisition utility per MC sample based on objective value obj.
        Should utilize the result of `set_X_pending` as needed to account for pending
        function evaluations.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of acquisition utility values.
        """
        pass  # pragma: no cover

    def _apply_constraints(self, acqval: Tensor, samples: Tensor) -> Tensor:
        """Multiplies the acquisition utility by constraint indicators.

        Args:
            acqval: `sample_shape x batch_shape x q`-dim acquisition utility values.
            samples: `sample_shape x batch_shape x q x m`-dim posterior samples.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of acquisition utility values
                multiplied by a smoothed constraint indicator per sample.
        """
        if self._constraints is not None:
            if not self._log and (acqval < 0).any():
                raise ValueError(
                    "Constraint-weighting requires unconstrained "
                    "acquisition values to be non-negative."
                )
            ind = compute_smoothed_feasibility_indicator(
                constraints=self._constraints,
                samples=samples,
                eta=self._eta,
                log=self._log,
                fat=self._fat,
            )
            acqval = acqval.add(ind) if self._log else acqval.mul(ind)
        return acqval


class qExpectedImprovement(SampleReducingMCAcquisitionFunction):
    r"""MC-based batch Expected Improvement.

    This computes qEI by
    (1) sampling the joint posterior over q points
    (2) evaluating the improvement over the current best for each sample
    (3) maximizing over q
    (4) averaging over the samples

    `qEI(X) = E(max(max Y - best_f, 0)), Y ~ f(X), where X = (x_1,...,x_q)`

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> best_f = train_Y.max()[0]
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qEI = qExpectedImprovement(model, best_f, sampler)
        >>> qei = qEI(test_X)

    NOTE: It is strongly recommended to use qLogExpectedImprovement instead
    of regular qEI, as it can lead to substantially improved BO performance through
    improved numerics. See https://arxiv.org/abs/2310.20708 for details.
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
    ) -> None:
        r"""q-Expected Improvement.

        Args:
            model: A fitted model.
            best_f: The best objective value observed so far (assumed noiseless). Can be
                a scalar, or a `batch_shape`-dim tensor. In case of a batched model, the
                tensor can specify different values for each element of the batch.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are evaluated.
                Defaults to `IdentityMCObjective()`.
                NOTE: `ConstrainedMCObjective` for outcome constraints is deprecated in
                favor of passing the `constraints` directly to this constructor.
            posterior_transform: A PosteriorTransform (optional).
            X_pending:  A `m x d`-dim Tensor of `m` design points that have been
                submitted for function evaluation but have not yet been evaluated.
                Concatenated into X upon forward call. Copied and set to have no
                gradient.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are considered satisfied if the output is less than zero.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. For more details, on this
                parameter, see the docs of `compute_smoothed_feasibility_indicator`.
        """
        legacy_ei_numerics_warning(legacy_name=type(self).__name__)
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            constraints=constraints,
            eta=eta,
        )
        self.register_buffer("best_f", torch.as_tensor(best_f))

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qExpectedImprovement per sample on the candidate set `X`.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of improvement utility values.
        """
        return (obj - self.best_f.unsqueeze(-1).to(obj)).clamp_min(0)


class qNoisyExpectedImprovement(
    SampleReducingMCAcquisitionFunction, CachedCholeskyMCSamplerMixin
):
    r"""MC-based batch Noisy Expected Improvement.

    This function does not assume a `best_f` is known (which would require
    noiseless observations). Instead, it uses samples from the joint posterior
    over the `q` test points and previously observed points. The improvement
    over previously observed points is computed for each sample and averaged.

    `qNEI(X) = E(max(max Y - max Y_baseline, 0))`, where
    `(Y, Y_baseline) ~ f((X, X_baseline)), X = (x_1,...,x_q)`

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qNEI = qNoisyExpectedImprovement(model, train_X, sampler)
        >>> qnei = qNEI(test_X)

    NOTE: It is strongly recommended to use qLogNoisyExpectedImprovement instead
    of regular qNEI, as it can lead to substantially improved BO performance through
    improved numerics. See https://arxiv.org/abs/2310.20708 for details.
    """

    def __init__(
        self,
        model: Model,
        X_baseline: Tensor,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        prune_baseline: bool = True,
        cache_root: bool = True,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
        marginalize_dim: int | None = None,
    ) -> None:
        r"""q-Noisy Expected Improvement.

        Args:
            model: A fitted model.
            X_baseline: A `batch_shape x r x d`-dim Tensor of `r` design points
                that have already been observed. These points are considered as
                the potential best design point.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
                NOTE: `ConstrainedMCObjective` for outcome constraints is deprecated in
                favor of passing the `constraints` directly to this constructor.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` design points
                that have points that have been submitted for function evaluation
                but have not yet been evaluated. Concatenated into `X` upon
                forward call. Copied and set to have no gradient.
            prune_baseline: If True, remove points in `X_baseline` that are
                highly unlikely to be the best point. This can significantly
                improve performance and is generally recommended. In order to
                customize pruning parameters, instead manually call
                `botorch.acquisition.utils.prune_inferior_points` on `X_baseline`
                before instantiating the acquisition function.
            cache_root: A boolean indicating whether to cache the root
                decomposition over `X_baseline` and use low-rank updates.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are considered satisfied if the output is less than zero.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. For more details, on this
                parameter, see the docs of `compute_smoothed_feasibility_indicator`.
            marginalize_dim: The dimension to marginalize over.

        TODO: similar to qNEHVI, when we are using sequential greedy candidate
        selection, we could incorporate pending points X_baseline and compute
        the incremental qNEI from the new point. This would greatly increase
        efficiency for large batches.
        """
        legacy_ei_numerics_warning(legacy_name=type(self).__name__)
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            constraints=constraints,
            eta=eta,
        )
        CachedCholeskyMCSamplerMixin.__init__(
            self, model=model, cache_root=cache_root, sampler=sampler
        )
        if prune_baseline:
            X_baseline = prune_inferior_points(
                model=model,
                X=X_baseline,
                objective=objective,
                posterior_transform=posterior_transform,
                constraints=self._constraints,
                marginalize_dim=marginalize_dim,
            )
        self.register_buffer("X_baseline", X_baseline)
        # registering buffers for _get_samples_and_objectives in the next `if` block
        self.register_buffer("baseline_samples", None)
        self.register_buffer("baseline_obj", None)
        if self._cache_root:
            self.q_in = -1
            # set baseline samples
            with torch.no_grad():  # this is _get_samples_and_objectives(X_baseline)
                posterior = self.model.posterior(
                    X_baseline, posterior_transform=self.posterior_transform
                )
                # Note: The root decomposition is cached in two different places. It
                # may be confusing to have two different caches, but this is not
                # trivial to change since each is needed for a different reason:
                # - LinearOperator caching to `posterior.mvn` allows for reuse within
                #  this function, which may be helpful if the same root decomposition
                #  is produced by the calls to `self.base_sampler` and
                #  `self._cache_root_decomposition`.
                # - self._baseline_L allows a root decomposition to be persisted outside
                #   this method.
                baseline_samples = self.get_posterior_samples(posterior)
                baseline_obj = self.objective(baseline_samples, X=X_baseline)

            # We make a copy here because we will write an attribute `base_samples`
            # to `self.base_sampler.base_samples`, and we don't want to mutate
            # `self.sampler`.
            self.base_sampler = deepcopy(self.sampler)
            self.baseline_samples = baseline_samples
            self.baseline_obj = baseline_obj
            self.register_buffer(
                "_baseline_best_f",
                self._compute_best_feasible_objective(
                    samples=baseline_samples, obj=baseline_obj
                ),  # `sample_shape x batch_shape`-dim
            )
            self._baseline_L = self._compute_root_decomposition(posterior=posterior)

    def compute_best_f(self, obj: Tensor) -> Tensor:
        """Computes the best (feasible) noisy objective value.

        Args:
            obj: `sample_shape x batch_shape x q`-dim Tensor of objectives in forward.

        Returns:
            A `sample_shape x batch_shape`-dim Tensor of best feasible objectives.
        """
        if self._cache_root:
            val = self._baseline_best_f
        else:
            val = self._compute_best_feasible_objective(
                samples=self.baseline_samples, obj=self.baseline_obj
            )
        # ensuring shape, dtype, device compatibility with obj
        n_sample_dims = len(self.sample_shape)
        view_shape = torch.Size(
            [
                *val.shape[:n_sample_dims],  # sample dimensions
                *(1,) * (obj.ndim - val.ndim - 1),  # pad to match obj, without `q`-dim
                *val.shape[n_sample_dims:],  # the rest
            ]
        )
        return val.view(view_shape).to(obj)  # obj.shape[:-1], i.e. without `q`-dim`

    def _sample_forward(self, obj: Tensor) -> Tensor:
        """Evaluate qNoisyExpectedImprovement per objective value in `obj`.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of noisy improvement values.
        """
        return (obj - self.compute_best_f(obj).unsqueeze(-1)).clamp_min(0)

    def _get_samples_and_objectives(self, X: Tensor) -> tuple[Tensor, Tensor]:
        r"""Compute samples at new points, using the cached root decomposition.

        Args:
            X: A `batch_shape x q x d`-dim tensor of inputs.

        Returns:
            A two-tuple `(samples, obj)`, where `samples` is a tensor of posterior
            samples with shape `sample_shape x batch_shape x q x m`, and `obj` is a
            tensor of MC objective values with shape `sample_shape x batch_shape x q`.
        """
        q = X.shape[-2]
        X_full = torch.cat([match_batch_shape(self.X_baseline, X), X], dim=-2)
        # TODO: Implement more efficient way to compute posterior over both training and
        # test points in GPyTorch (https://github.com/cornellius-gp/gpytorch/issues/567)
        posterior = self.model.posterior(
            X_full, posterior_transform=self.posterior_transform
        )
        if not self._cache_root:
            samples_full = super().get_posterior_samples(posterior)
            samples = samples_full[..., -q:, :]
            obj_full = self.objective(samples_full, X=X_full)
            # assigning baseline buffers so `best_f` can be computed in _sample_forward
            self.baseline_obj, obj = obj_full[..., :-q], obj_full[..., -q:]
            self.baseline_samples = samples_full[..., :-q, :]
        else:
            # handle one-to-many input transforms
            n_plus_q = X_full.shape[-2]
            n_w = posterior._extended_shape()[-2] // n_plus_q
            q_in = q * n_w
            self._set_sampler(q_in=q_in, posterior=posterior)
            samples = self._get_f_X_samples(posterior=posterior, q_in=q_in)
            obj = self.objective(samples, X=X_full[..., -q:, :])

        return samples, obj

    def _compute_best_feasible_objective(self, samples: Tensor, obj: Tensor) -> Tensor:
        r"""Computes best feasible objective value from samples.

        Args:
            samples: `sample_shape x batch_shape x q x m`-dim posterior samples.
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape`-dim Tensor of best feasible objectives.
        """
        return compute_best_feasible_objective(
            samples=samples,
            obj=obj,
            constraints=self._constraints,
            model=self.model,
            objective=self.objective,
            posterior_transform=self.posterior_transform,
            X_baseline=self.X_baseline,
        )


class qProbabilityOfImprovement(SampleReducingMCAcquisitionFunction):
    r"""MC-based batch Probability of Improvement.

    Estimates the probability of improvement over the current best observed
    value by sampling from the joint posterior distribution of the q-batch.
    MC-based estimates of a probability involves taking expectation of an
    indicator function; to support auto-differentiation, the indicator is
    replaced with a sigmoid function with temperature parameter `tau`.

    `qPI(X) = P(max Y >= best_f), Y ~ f(X), X = (x_1,...,x_q)`

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> best_f = train_Y.max()[0]
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qPI = qProbabilityOfImprovement(model, best_f, sampler)
        >>> qpi = qPI(test_X)
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        tau: float = 1e-3,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
    ) -> None:
        r"""q-Probability of Improvement.

        Args:
            model: A fitted model.
            best_f: The best objective value observed so far (assumed noiseless). Can
                be a `batch_shape`-shaped tensor, which in case of a batched model
                specifies potentially different values for each element of the batch.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
                NOTE: `ConstrainedMCObjective` for outcome constraints is deprecated in
                favor of passing the `constraints` directly to this constructor.
            posterior_transform: A PosteriorTransform (optional).
            X_pending:  A `m x d`-dim Tensor of `m` design points that have
                points that have been submitted for function evaluation
                but have not yet been evaluated.  Concatenated into X upon
                forward call.  Copied and set to have no gradient.
            tau: The temperature parameter used in the sigmoid approximation
                of the step function. Smaller values yield more accurate
                approximations of the function, but result in gradients
                estimates with higher variance.
            constraints: A list of constraint callables which map posterior samples to
                a scalar. The associated constraint is considered satisfied if this
                scalar is less than zero.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. For more details, on this
                parameter, see the docs of `compute_smoothed_feasibility_indicator`.
        """
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            constraints=constraints,
            eta=eta,
        )
        best_f = torch.as_tensor(best_f).unsqueeze(-1)  # adding batch dim
        self.register_buffer("best_f", best_f)
        self.register_buffer("tau", torch.as_tensor(tau))

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qProbabilityOfImprovement per sample on the candidate set `X`.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of improvement indicators.
        """
        improvement = obj - self.best_f.to(obj)
        return torch.sigmoid(improvement / self.tau)


class qSimpleRegret(SampleReducingMCAcquisitionFunction):
    r"""MC-based batch Simple Regret.

    Samples from the joint posterior over the q-batch and computes the simple regret.

    `qSR(X) = E(max Y), Y ~ f(X), X = (x_1,...,x_q)`

    Constraints should be provided as a `ConstrainedMCObjective`.
    Passing `constraints` as an argument is not supported. This is because
    `SampleReducingMCAcquisitionFunction` computes the acquisition values on the sample
    level and then weights the sample-level acquisition values by a soft feasibility
    indicator. Hence, it expects non-log acquisition function values to be
    non-negative. `qSimpleRegret` acquisition values can be negative, so we instead use
    a `ConstrainedMCObjective` which applies constraints to the objectives (e.g. before
    computing the acquisition function) and shifts negative objective values using
    an infeasible cost to ensure non-negativity (before applying constraints and
    shifting them back).

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qSR = qSimpleRegret(model, sampler)
        >>> qsr = qSR(test_X)
    """

    def __init__(
        self,
        model: Model,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
    ) -> None:
        r"""q-Simple Regret.

        Args:
            model: A fitted model.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending:  A `m x d`-dim Tensor of `m` design points that have
                points that have been submitted for function evaluation
                but have not yet been evaluated.  Concatenated into X upon
                forward call.  Copied and set to have no gradient.
        """
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
        )

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qSimpleRegret per sample on the candidate set `X`.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of simple regret values.
        """
        return obj


class qUpperConfidenceBound(SampleReducingMCAcquisitionFunction):
    r"""MC-based batch Upper Confidence Bound.

    Uses a reparameterization to extend UCB to qUCB for q > 1 (See Appendix A
    of [Wilson2017reparam].)

    `qUCB = E(max(mu + |Y_tilde - mu|))`, where `Y_tilde ~ N(mu, beta pi/2 Sigma)`
    and `f(X)` has distribution `N(mu, Sigma)`.

    Constraints should be provided as a `ConstrainedMCObjective`.
    Passing `constraints` as an argument is not supported. This is because
    `SampleReducingMCAcquisitionFunction` computes the acquisition values on the sample
    level and then weights the sample-level acquisition values by a soft feasibility
    indicator. Hence, it expects non-log acquisition function values to be
    non-negative. `qUpperConfidenceBound` acquisition values can be negative, so we
    instead use a `ConstrainedMCObjective` which applies constraints to the objectives
    (e.g. before computing the acquisition function) and shifts negative objective
    values using an infeasible cost to ensure non-negativity (before applying
    constraints and shifting them back).

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qUCB = qUpperConfidenceBound(model, 0.1, sampler)
        >>> qucb = qUCB(test_X)
    """

    def __init__(
        self,
        model: Model,
        beta: float,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
    ) -> None:
        r"""q-Upper Confidence Bound.

        Args:
            model: A fitted model.
            beta: Controls tradeoff between mean and standard deviation in UCB.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` design points that have
                points that have been submitted for function evaluation but have not yet
                been evaluated. Concatenated into X upon forward call. Copied and set to
                have no gradient.
        """
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
        )
        self.beta_prime = self._get_beta_prime(beta=beta)

    def _get_beta_prime(self, beta: float) -> float:
        return math.sqrt(beta * math.pi / 2)

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qUpperConfidenceBound per sample on the candidate set `X`.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of acquisition values.
        """
        mean = obj.mean(dim=0)
        return mean + self.beta_prime * (obj - mean).abs()


class qLowerConfidenceBound(qUpperConfidenceBound):
    r"""MC-based batched lower confidence bound.

    This acquisition function is useful for confident/risk-averse decision making.
    This acquisition function is intended to be maximized as with qUpperConfidenceBound,
    but the qLowerConfidenceBound will be pessimistic in the face of uncertainty and
    lead to conservative candidates.
    """

    def _get_beta_prime(self, beta: float) -> float:
        """Multiply beta prime by -1 to get the lower confidence bound."""
        return -super()._get_beta_prime(beta=beta)


class qPosteriorStandardDeviation(SampleReducingMCAcquisitionFunction):
    r"""MC-based batch Posterior Standard Deviation.

    An acquisition function for pure exploration.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qPSTD = qPosteriorStandardDeviation(model, sampler)
        >>> std = qPSTD(test_X)
    """

    def __init__(
        self,
        model: Model,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
    ) -> None:
        r"""q-Posterior Standard Deviation.

        Args:
            model: A fitted model.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` design points that have
                points that have been submitted for function evaluation but have not yet
                been evaluated. Concatenated into X upon forward call. Copied and set to
                have no gradient.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are considered satisfied if the output is less than zero.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. For more details, on this
                parameter, see the docs of `compute_smoothed_feasibility_indicator`.
        """
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            constraints=constraints,
            eta=eta,
        )
        self._scale = math.sqrt(math.pi / 2)

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qPosteriorStandardDeviation per sample on the candidate set `X`.

        Args:
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of acquisition values.
        """
        mean = obj.mean(dim=0)
        return (obj - mean).abs() * self._scale



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
This module defines the botorch model for PFNs (`PFNModel`) and it
provides handy helpers to download pretrained, public PFNs
with `download_model` and model paths with `ModelPaths`.
For the latter to work `pfns4bo` must be installed.
"""

from __future__ import annotations

from typing import Any, Optional, Union

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import UnsupportedError
from botorch.logging import logger
from botorch.models.model import Model
from botorch.models.transforms.input import InputTransform
from botorch.utils.transforms import match_batch_shape
from botorch_community.models.utils.prior_fitted_network import (
    download_model,
    ModelPaths,
)
from botorch_community.posteriors.riemann import (
    BoundedRiemannPosterior,
    MultivariateRiemannPosterior,
)
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from pfns.train import MainConfig  # @manual=//pytorch/PFNs:PFNs
from torch import Tensor
from torch.nn import Module


class PFNModel(Model):
    """Prior-data Fitted Network"""

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        model: Module | None = None,
        checkpoint_url: str = ModelPaths.pfns4bo_hebo,
        train_Yvar: Tensor | None = None,
        batch_first: bool = False,
        constant_model_kwargs: dict[str, Any] | None = None,
        input_transform: InputTransform | None = None,
        load_training_checkpoint: bool = False,
    ) -> None:
        """Initialize a PFNModel.

        Either a pre-trained PFN model can be provided via the model kwarg,
        or a checkpoint_url can be provided from which the model will be
        downloaded. This defaults to the pfns4bo_hebo model.

        Loading the model does an unsafe "weights_only=False" load, so
        it is essential that checkpoint_url be a trusted source.

        Args:
            train_X: A `n x d` tensor of training features.
            train_Y: A `n x 1` tensor of training observations.
            model: A pre-trained PFN model with the following
                forward(train_X, train_Y, X) -> logit predictions of shape
                `n x b x c` where c is the number of discrete buckets
                borders: A `c+1`-dim tensor of bucket borders.
            checkpoint_url: The string URL of the PFN model to download and load.
                Will be ignored if model is provided.
            train_Yvar: Observed variance of train_Y. Currently ignored.
            batch_first: Whether the batch dimension is the first dimension of
                the input tensors. This is needed to support different PFN
                models. For batch-first x has shape `batch x seq_len x features`
                and for non-batch-first it has shape `seq_len x batch x features`.
            constant_model_kwargs: A dictionary of model kwargs that
                will be passed to the model in each forward pass.
            input_transform: A Botorch input transform.
            load_training_checkpoint: Whether to load a training checkpoint as
                produced by the PFNs training code, see github.com/automl/PFNs.

        """
        super().__init__()
        if model is None:
            model = download_model(
                model_path=checkpoint_url,
            )

        if load_training_checkpoint:
            # the model is not an actual model, but a training checkpoint
            # make a model out of it
            checkpoint = model
            config = MainConfig.from_dict(checkpoint["config"])
            model = config.model.create_model()
            model.load_state_dict(checkpoint["model_state_dict"])
            model.eval()

        if train_Yvar is not None:
            logger.debug("train_Yvar provided but ignored for PFNModel.")

        if train_Y.dim() != 2:
            raise UnsupportedError("train_Y must be 2-dimensional.")

        if train_X.dim() != 2:
            raise UnsupportedError("train_X must be 2-dimensional.")

        if train_Y.shape[-1] > 1:
            raise UnsupportedError("Only 1 target allowed for PFNModel.")

        if train_X.shape[0] != train_Y.shape[0]:
            raise UnsupportedError(
                "train_X and train_Y must have the same number of rows."
            )

        with torch.no_grad():
            self.transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )

        self.train_X = train_X  # shape: (n, d)
        self.train_Y = train_Y  # shape: (n, 1)
        # Downstream botorch tooling expects a likelihood to be specified,
        # so here we use a FixedNoiseGaussianLikelihood that is unused.
        if train_Yvar is None:
            train_Yvar = torch.zeros_like(train_Y)
        self.likelihood = FixedNoiseGaussianLikelihood(noise=train_Yvar)
        self.pfn = model.to(device=train_X.device)
        self.batch_first = batch_first
        self.constant_model_kwargs = constant_model_kwargs or {}
        if input_transform is not None:
            self.input_transform = input_transform

    def posterior(
        self,
        X: Tensor,
        output_indices: Optional[list[int]] = None,
        observation_noise: Union[bool, Tensor] = False,
        posterior_transform: Optional[PosteriorTransform] = None,
    ) -> BoundedRiemannPosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Note: The input transforms should be applied here using
            `self.transform_inputs(X)` after the `self.eval()` call and before
            any `model.forward` or `model.likelihood` calls.

        Args:
            X: A b? x q? x d`-dim Tensor, where `d` is the dimension of the
                feature space.
            output_indices: **Currently not supported for PFNModel.**
            observation_noise: **Currently not supported for PFNModel**.
            posterior_transform: **Currently not supported for PFNModel**.

        Returns:
            A `BoundedRiemannPosterior`, representing a batch of b? x q?`
            distributions.
        """
        self.pfn.eval()
        if output_indices is not None:
            raise UnsupportedError(
                "output_indices is not None. PFNModel should not "
                "be a multi-output model."
            )
        if observation_noise:
            logger.warning(
                "observation_noise is not supported for PFNModel and is being ignored."
            )
        if posterior_transform is not None:
            raise UnsupportedError("posterior_transform is not supported for PFNModel.")

        X, train_X, train_Y, orig_X_shape = self._prepare_data(X)

        probabilities = self.pfn_predict(
            X=X, train_X=train_X, train_Y=train_Y
        )  # (b, q, num_buckets)
        probabilities = probabilities.view(
            *orig_X_shape[:-1], -1
        )  # (b?, q?, num_buckets)

        return BoundedRiemannPosterior(
            borders=self.borders,
            probabilities=probabilities,
        )

    def _prepare_data(self, X: Tensor) -> tuple[Tensor, Tensor, Tensor, torch.Size]:
        orig_X_shape = X.shape  # X has shape b? x q? x d
        if len(X.shape) > 3:
            raise UnsupportedError(f"X must be at most 3-d, got {X.shape}.")
        while len(X.shape) < 3:
            X = X.unsqueeze(0)

        X = self.transform_inputs(X)  # shape (b , q, d)

        train_X = match_batch_shape(self.transformed_X, X)  # shape (b, n, d)
        train_Y = match_batch_shape(self.train_Y, X)  # shape (b, n, 1)
        return X, train_X, train_Y, orig_X_shape

    def pfn_predict(self, X: Tensor, train_X: Tensor, train_Y: Tensor) -> Tensor:
        """
        Make a prediction using the PFN model on X given training data.

        Args:
            X: has shape (b, q, d)
            train_X: has shape (b, n, d)
            train_Y: has shape (b, n, 1)

        Returns: probabilities (b, q, num_buckets) for Riemann posterior.
        """
        if not self.batch_first:
            X = X.transpose(0, 1)  # shape (q, b, d)
            train_X = train_X.transpose(0, 1)  # shape (n, b, d)
            train_Y = train_Y.transpose(0, 1)  # shape (n, b, 1)

        logits = self.pfn(
            train_X.float(),
            train_Y.float(),
            X.float(),
            **self.constant_model_kwargs,
        )
        if not self.batch_first:
            logits = logits.transpose(0, 1)  # shape (b, q, num_buckets)
        logits = logits.to(X.dtype)

        probabilities = logits.softmax(dim=-1)  # shape (b, q, num_buckets)
        return probabilities

    @property
    def borders(self):
        return self.pfn.criterion.borders.to(self.train_X.dtype)


class MultivariatePFNModel(PFNModel):
    """A multivariate PFN model that returns a joint posterior over q batch inputs.

    For this to work correctly it is necessary that the underlying model return a
    posterior for the latent f, not the noisy observed y.
    """

    def posterior(
        self,
        X: Tensor,
        output_indices: Optional[list[int]] = None,
        observation_noise: Union[bool, Tensor] = False,
        posterior_transform: Optional[PosteriorTransform] = None,
    ) -> Union[BoundedRiemannPosterior, MultivariateRiemannPosterior]:
        """Computes the posterior over model outputs at the provided points.

        Will produce a MultivariateRiemannPosterior that fits a joint structure
        over the q batch dimension of X. This will require an additional forward
        pass through the PFN model, and some approximation.

        If q = 1 or there is no q dimension, will return a BoundedRiemannPosterior
        and behave the same as PFNModel.

        Args:
            X: A b? x q? x d`-dim Tensor, where `d` is the dimension of the
                feature space.
            output_indices: **Currently not supported for PFNModel.**
            observation_noise: **Currently not supported for PFNModel**.
            posterior_transform: **Currently not supported for PFNModel**.

        Returns:
            A posterior representing a batch of b? x q? distributions.
        """
        marginals = super().posterior(
            X=X,
            output_indices=output_indices,
            observation_noise=observation_noise,
            posterior_transform=posterior_transform,
        )
        if len(X.shape) == 1 or X.shape[-2] == 1:
            # No q dimension, or q=1
            return marginals
        X, train_X, train_Y, orig_X_shape = self._prepare_data(X)
        # Estimate correlation structure, making another forward pass.
        R = self.estimate_correlations(
            X=X,
            train_X=train_X,
            train_Y=train_Y,
            marginals=marginals,
        )  # (b, q, q)
        R = R.view(*orig_X_shape[:-2], X.shape[-2], X.shape[-2])  # (b?, q, q)
        return MultivariateRiemannPosterior(
            borders=self.borders,
            probabilities=marginals.probabilities,
            correlation_matrix=R,
        )

    def estimate_correlations(
        self,
        X: Tensor,
        train_X: Tensor,
        train_Y: Tensor,
        marginals: BoundedRiemannPosterior,
    ) -> Tensor:
        """
        Estimate a correlation matrix R across the q batch of points in X.
        Will do a forward pass through the PFN model with batch size O(q^2).

        For every x_q in [x_1, ..., x_Q]:
           1. Add x_q to train_X, with y_q the 90th percentile value for f(x_q)
           2. Evaluate p(f(x_i)) for all points.

        Uses bivariate normal conditioning formulae, and so will be approximate.

        Args:
            X: evaluation point, shape (b, q, d)
            train_X: Training X, shape (b, n, d)
            train_Y: Training Y, shape (b, n, 1)
            marginals: A posterior object with marginal posteriors for f(X), but no
                correlation structure yet added. posterior.probabilities has
                shape (b?, q, num_buckets).

        Returns: A (b, q, q) correlation matrix
        """
        # Compute conditional distributions with a forward pass
        cond_mean, cond_val = self._compute_conditional_means(
            X=X,
            train_X=train_X,
            train_Y=train_Y,
            marginals=marginals,
        )
        # Get marginal moments
        var = marginals.variance.squeeze(-1)  # (b?, q)
        mean = marginals.mean.squeeze(-1)  # (b?, q)
        if len(var.shape) == 1:
            var = var.unsqueeze(0)  # (b, q)
            mean = mean.unsqueeze(0)  # (b, q)
        # Estimate covariances from conditional distributions
        cov = self._estimate_covariances(
            cond_mean=cond_mean,
            cond_val=cond_val,
            mean=mean,
            var=var,
        )
        # Convert to correlation matrix
        S = 1 / torch.sqrt(torch.diagonal(cov, dim1=-2, dim2=-1))  # (b, q)
        S = S.unsqueeze(-1).expand(cov.shape)  # (b, q, q)
        R = S * cov * S.transpose(-1, -2)  # (b, q, q)
        return R

    def _compute_conditional_means(
        self,
        X: Tensor,
        train_X: Tensor,
        train_Y: Tensor,
        marginals: BoundedRiemannPosterior,
    ) -> tuple[Tensor, Tensor]:
        """
        Compute conditional means between pairs of points in X.

        Conditioning is done with an additional forward pass through the model. The
        returned conditional mean will be of shape (b, q, q), with entry [b, i, j] the
        conditional mean of j given i set to the conditioning value.

        Args:
            X: evaluation point, shape (b, q, d)
            train_X: Training X, shape (b, n, d)
            train_Y: Training Y, shape (b, n, 1)
            marginals: A posterior object with marginal posteriors for f(X), but no
                correlation structure yet added. posterior.probabilities has
                shape (b?, q, num_buckets).

        Returns: conditional means (b, q, q), and values used for conditioning (b, q).
        """
        b, q, d = X.shape
        n = train_X.shape[-2]
        post_shape = marginals.probabilities.shape[:-1]
        # Find the 90th percentile of each eval point.
        cond_val = marginals.icdf(
            torch.full(post_shape, 0.9, device=X.device, dtype=X.dtype).unsqueeze(0)
        )  # (1, b?, q, 1)
        cond_val = cond_val.view(b, q)  # (b, q)
        # Construct conditional training data.
        # train_X will have shape (b, q, n+1, d), to have a conditional observation
        # for each point. train_Y will have shape (b, q, n+1, 1).
        train_X = train_X.unsqueeze(1).expand(b, q, n, d)
        cond_X = X.unsqueeze(-2)  # (b, q, 1, d)
        train_X = torch.cat((train_X, cond_X), dim=-2)  # (b, q, n+1, d)
        train_Y = train_Y.unsqueeze(1).expand(b, q, n, 1)
        cond_Y = cond_val.unsqueeze(-1).unsqueeze(-1)  # (b, q, 1, 1)
        train_Y = torch.cat((train_Y, cond_Y), dim=-2)  # (b, q, n+1, 1)
        # Construct eval points
        eval_X = X.unsqueeze(1).expand(b, q, q, d)
        # Squeeze everything into necessary 2 batch dims, and do PFN forward pass
        cond_probabilities = self.pfn_predict(
            X=eval_X.reshape(b * q, q, d),
            train_X=train_X.reshape(b * q, n + 1, d),
            train_Y=train_Y.reshape(b * q, n + 1, 1),
        )  # (b * q, q, num_buckets)
        # Object for conditional posteriors
        cond_posterior = BoundedRiemannPosterior(
            borders=self.borders,
            probabilities=cond_probabilities,
        )
        # Get conditional means
        cond_mean = cond_posterior.mean.squeeze(-1)  # (b * q, q)
        cond_mean = cond_mean.unsqueeze(0).view(b, q, q)
        return cond_mean, cond_val

    def _estimate_covariances(
        self,
        cond_mean: Tensor,
        cond_val: Tensor,
        mean: Tensor,
        var: Tensor,
    ) -> Tensor:
        """
        Estimate covariances from conditional distributions.

        Part one: Compute noise variance implied by conditional distributions
        E[f_j | y_j=y] = E[f_j] + var[f_j]/(var[f_j] + noise_var) * (y - E[f_j])
        Let Z_jj = (E[f_j | y_j=y] - E[f_j]) / (y - E[f_j]).
        Note that Z is in (0, 1].
        Then, noise_var_j = var[f_j](1/Z_jj - 1).

        Part two: Compute covariances for all pairs
        E[f_j|y_i=y] = E[f_j]+cov[f_j, f_i]/(var[f_i] + noise_var_i) * (y - E[f_i])
        Let Z_ij = (E[f_j | y_i=y] - E[f_j]) / (y - E[f_i]).
        Then, cov[f_j, f_i] = Z * (var[f_i] + noise_var)

        Args:
            cond_mean: (b, q, q) means of dim -1 conditioned on dim -2
            cond_val: (b, q) conditioned y value.
            var: (b, q) marginal variances
            mean: (b, q) marginal means

        Returns: Covariance matrix
        """
        Z = (cond_mean - mean.unsqueeze(-2).expand(cond_mean.shape)) / (
            cond_val - mean
        ).unsqueeze(-1)  # (b, q, q)
        # Z[i, j] is for j cond. on i
        noise_var = torch.clamp(
            var * (1 / torch.diagonal(Z, dim1=-2, dim2=-1) - 1), min=1e-8
        )  # (b, q)
        cov = Z * (var + noise_var).unsqueeze(-1)  # (b, q, q)
        # Symmetrize
        cov = 0.5 * (cov + cov.transpose(-1, -2))
        cov = self._map_psd(cov)
        return cov

    def _map_psd(self, A):
        """
        Map A (assumed symmetric) to the nearest PSD matrix.
        """
        if torch.linalg.eigvals(A).real.min() < 0:
            L, Q = torch.linalg.eigh(A)
            L = torch.clamp(L, min=1e-6)
            A = Q @ torch.diag_embed(L) @ Q.transpose(-1, -2)
        return A



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
from typing import Any

import torch
from botorch.acquisition.objective import ScalarizedPosteriorTransform
from botorch.exceptions import BotorchWarning, UnsupportedError
from botorch.models import SingleTaskGP
from botorch.models.transforms import Normalize
from botorch.posteriors import GPyTorchPosterior
from botorch.sampling.normal import IIDNormalSampler, SobolQMCNormalSampler
from botorch.utils.testing import BotorchTestCase, MockModel, MockPosterior
from botorch_community.acquisition.rei import (
    LogRegionalExpectedImprovement,
    qLogRegionalExpectedImprovement,
)
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal


class TestLogRegionalExpectedImprovement(BotorchTestCase):
    def test_log_regional_expected_improvement(self):
        for dtype in (torch.float, torch.double):
            self._test_log_regional_expected_improvement(dtype=dtype)

    def _test_log_regional_expected_improvement(self, dtype: torch.dtype) -> None:
        mean = torch.tensor([[-0.5]], device=self.device, dtype=dtype)
        variance = torch.ones(1, 1, device=self.device, dtype=dtype)
        mm = MockModel(MockPosterior(mean=mean, variance=variance))

        # basic test
        X_dev = torch.rand(1, 1, device=self.device, dtype=dtype)
        module = LogRegionalExpectedImprovement(model=mm, best_f=0.0, X_dev=X_dev)
        X = torch.empty(1, 1, device=self.device, dtype=dtype)
        log_rei = module(X)
        rei_expected = torch.tensor([0.19780], device=self.device, dtype=dtype)
        self.assertAllClose(log_rei, rei_expected.log(), atol=1e-4)

        # test maximize
        X_dev = torch.rand(1, 1, device=self.device, dtype=dtype)
        module = LogRegionalExpectedImprovement(
            model=mm, best_f=0.0, X_dev=X_dev, maximize=False
        )
        X = torch.empty(1, 1, device=self.device, dtype=dtype)
        log_rei = module(X)
        rei_expected = torch.tensor([0.6978], device=self.device, dtype=dtype)
        self.assertAllClose(log_rei, rei_expected.log(), atol=1e-4)

        with self.assertRaisesRegex(
            UnsupportedError,
            "Analytic acquisition functions do not account for X_pending yet.",
        ):
            module.set_X_pending(None)

        # test bounds argument
        X_dev = torch.rand(1, 1, device=self.device, dtype=dtype)
        bounds = torch.tensor([[0.0], [1.0]], device=self.device, dtype=dtype)
        module = LogRegionalExpectedImprovement(
            model=mm, best_f=0.0, X_dev=X_dev, bounds=bounds
        )
        X = torch.empty(1, 1, device=self.device, dtype=dtype)
        log_rei = module(X)
        rei_expected = torch.tensor([0.19780], device=self.device, dtype=dtype)
        self.assertAllClose(log_rei, rei_expected.log(), atol=1e-4)

        # test posterior transform (single-output)
        mean = torch.tensor([0.5], device=self.device, dtype=dtype)
        covar = torch.tensor([[0.16]], device=self.device, dtype=dtype)
        mvn = MultivariateNormal(mean, covar)
        p = GPyTorchPosterior(mvn)
        mm = MockModel(p)
        weights = torch.tensor([0.5], device=self.device, dtype=dtype)
        transform = ScalarizedPosteriorTransform(weights)
        X_dev = torch.rand(1, 2, device=self.device, dtype=dtype)
        log_rei = LogRegionalExpectedImprovement(
            model=mm, best_f=0.0, X_dev=X_dev, posterior_transform=transform
        )
        X = torch.rand(1, 2, device=self.device, dtype=dtype)
        rei_expected = torch.tensor([0.2601], device=self.device, dtype=dtype)
        self.assertAllClose(log_rei(X), rei_expected.log(), atol=1e-4)

        # test posterior transform (multi-output)
        mean = torch.tensor([[-0.25, 0.5]], device=self.device, dtype=dtype)
        covar = torch.tensor(
            [[[0.5, 0.125], [0.125, 0.5]]], device=self.device, dtype=dtype
        )
        mvn = MultitaskMultivariateNormal(mean, covar)
        p = GPyTorchPosterior(mvn)
        mm = MockModel(p)
        weights = torch.tensor([2.0, 1.0], device=self.device, dtype=dtype)
        transform = ScalarizedPosteriorTransform(weights)
        X_dev = torch.rand(1, 2, device=self.device, dtype=dtype)
        log_rei = LogRegionalExpectedImprovement(
            model=mm, best_f=0.0, X_dev=X_dev, posterior_transform=transform
        )
        X = torch.rand(1, 2, device=self.device, dtype=dtype)
        rei_expected = torch.tensor([0.6910], device=self.device, dtype=dtype)
        self.assertAllClose(log_rei(X), rei_expected.log(), atol=1e-4)


class TestQLogRegionalExpectedImprovement(BotorchTestCase):
    def test_q_log_regional_expected_improvement(self):
        for dtype in (torch.float, torch.double):
            with self.subTest(dtype=dtype):
                self._test_q_log_regional_expected_improvement(dtype)

    def _test_q_log_regional_expected_improvement(self, dtype: torch.dtype) -> None:
        tkwargs: dict[str, Any] = {"device": self.device, "dtype": dtype}
        # `mc_model_samples x mc_X_dev_samples x q x d` = 1 x 1 x 1 x 1
        samples = torch.zeros(1, 1, 1, 1, **tkwargs)
        mm = MockModel(MockPosterior(samples=samples))
        # X is `q x d` = 1 x 1. X is a dummy and unused b/c of mocking
        X = torch.zeros(1, 1, **tkwargs)
        X_dev = torch.zeros(1, 1, **tkwargs)
        bounds = torch.tensor([[0.0], [1.0]], **tkwargs)

        # basic test
        sampler = IIDNormalSampler(sample_shape=torch.Size([2]))
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=0, X_dev=X_dev, sampler=sampler
        )
        # test initialization
        for k in ["objective", "sampler"]:
            self.assertIn(k, acqf._modules)

        res = acqf(X)
        # self.assertEqual(res.item(), 0.0)
        self.assertAlmostEqual(res.item(), -14.0473, places=4)

        # test shifting best_f value
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=-1, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertEqual(res.item(), 0.0)

        # test bounds argument
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=-1, X_dev=X_dev, sampler=sampler, bounds=bounds
        )
        res = acqf(X)
        self.assertEqual(res.item(), 0.0)

        # basic test, no resample
        sampler = IIDNormalSampler(sample_shape=torch.Size([2]), seed=12345)
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=0, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertAlmostEqual(res.item(), -14.0473, places=4)
        self.assertEqual(acqf.sampler.base_samples.shape, torch.Size([2, 1, 1, 1, 1]))
        bs = acqf.sampler.base_samples.clone()
        res = acqf(X)
        self.assertTrue(torch.equal(acqf.sampler.base_samples, bs))

        # basic test, qmc
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([2]))
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=0, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertAlmostEqual(res.item(), -14.0473, places=4)
        self.assertEqual(acqf.sampler.base_samples.shape, torch.Size([2, 1, 1, 1, 1]))
        bs = acqf.sampler.base_samples.clone()
        acqf(X)
        self.assertTrue(torch.equal(acqf.sampler.base_samples, bs))

        # basic test for X_pending and warning
        acqf.set_X_pending()
        self.assertIsNone(acqf.X_pending)
        acqf.set_X_pending(None)
        self.assertIsNone(acqf.X_pending)
        acqf.set_X_pending(X)
        self.assertEqual(acqf.X_pending, X)
        mm._posterior._samples = torch.zeros(1, 2, 1, **tkwargs)
        res = acqf(X)
        X2 = torch.zeros(1, 1, 1, **tkwargs, requires_grad=True)
        with warnings.catch_warnings(record=True) as ws:
            acqf.set_X_pending(X2)
        self.assertEqual(acqf.X_pending, X2)
        self.assertEqual(sum(issubclass(w.category, BotorchWarning) for w in ws), 1)

        # Test without mocking tensor dimensions
        n = 3
        d = 2
        X = torch.randn(n, d, dtype=torch.float64)
        Y = X.sin().sum(dim=-1, keepdim=True)

        model = SingleTaskGP(X, Y, input_transform=Normalize(d=X.shape[-1]))

        n_tr = 17
        X_dev = torch.rand(n_tr, d)
        acqf = qLogRegionalExpectedImprovement(model, best_f=0.0, X_dev=X_dev)
        batch_shape = ()
        X_test = torch.randn(*batch_shape, n, d)
        q_log_rei = acqf(X_test)
        self.assertEqual(q_log_rei.shape, torch.Size([1]))

    def test_q_log_regional_expected_improvement_batch(self):
        for dtype in (torch.float, torch.double):
            with self.subTest(dtype=dtype):
                self._test_q_log_regional_expected_improvement_batch(dtype)

    def _test_q_log_regional_expected_improvement_batch(
        self, dtype: torch.dtype
    ) -> None:
        # `mc_model_samples x mc_X_dev_samples x q x d` = 1 x 1 x 1 x 1
        samples = torch.zeros(2, 2, 2, 1, device=self.device, dtype=dtype)
        samples[0, 0, 0] = 1.0
        mm = MockModel(MockPosterior(samples=samples))

        # X is a dummy and unused b/c of mocking
        X = torch.zeros(2, 2, 1, device=self.device, dtype=dtype)
        X_dev = torch.zeros(2, 1, device=self.device, dtype=dtype)

        # test batch mode
        sampler = IIDNormalSampler(sample_shape=torch.Size([2]))
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=0, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertAlmostEqual(res[0].item(), -0.6931, places=4)
        self.assertAlmostEqual(res[1].item(), -14.0403, places=4)

        # test batch model, batched best_f values
        sampler = IIDNormalSampler(sample_shape=torch.Size([3]))
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=torch.Tensor([0, 0]), X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertAlmostEqual(res[0].item(), -0.6931, places=4)
        self.assertAlmostEqual(res[1].item(), -14.0403, places=4)

        # test shifting best_f value
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=-1, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertAlmostEqual(res[0].item(), 0.4078, places=4)
        self.assertAlmostEqual(res[1].item(), 0.0069, places=4)

        # test batch mode
        sampler = IIDNormalSampler(sample_shape=torch.Size([2]), seed=12345)
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=0, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)  # 1-dim batch
        self.assertAlmostEqual(res[0].item(), -0.6931, places=4)
        self.assertAlmostEqual(res[1].item(), -14.0403, places=4)
        self.assertEqual(acqf.sampler.base_samples.shape, torch.Size([2, 1, 1, 2, 1]))
        bs = acqf.sampler.base_samples.clone()
        acqf(X)
        self.assertTrue(torch.equal(acqf.sampler.base_samples, bs))
        res = acqf(X.expand(2, 2, 1))  # 2-dim batch
        self.assertAlmostEqual(res[0].item(), -0.6931, places=4)
        self.assertAlmostEqual(res[1].item(), -14.0403, places=4)
        # the base samples should have the batch dim collapsed
        self.assertEqual(acqf.sampler.base_samples.shape, torch.Size([2, 1, 1, 2, 1]))
        bs = acqf.sampler.base_samples.clone()
        acqf(X.expand(2, 2, 1))
        self.assertTrue(torch.equal(acqf.sampler.base_samples, bs))

        # test batch mode, qmc
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([2]))
        acqf = qLogRegionalExpectedImprovement(
            model=mm, best_f=0, X_dev=X_dev, sampler=sampler
        )
        res = acqf(X)
        self.assertAlmostEqual(res[0].item(), -0.6931, places=4)
        self.assertAlmostEqual(res[1].item(), -14.0403, places=4)
        self.assertEqual(acqf.sampler.base_samples.shape, torch.Size([2, 1, 1, 2, 1]))
        bs = acqf.sampler.base_samples.clone()
        acqf(X)
        self.assertTrue(torch.equal(acqf.sampler.base_samples, bs))

        # Test without mocking
        n = 3
        d = 2
        X = torch.randn(n, d, dtype=torch.float64)
        Y = X.sin().sum(dim=-1, keepdim=True)

        model = SingleTaskGP(
            X,
            Y,
            input_transform=Normalize(d=X.shape[-1]),
        )

        n_tr = 17
        X_dev = torch.rand(n_tr, d)
        acqf = qLogRegionalExpectedImprovement(model, best_f=0.0, X_dev=X_dev)
        # Test for non-trivial batch shape
        batch_shape = (5,)
        X_test = torch.randn(*batch_shape, n, d)
        q_log_rei = acqf(X_test)
        self.assertEqual(q_log_rei.shape, torch.Size([5]))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.models.approximate_gp import (
    ApproximateGPyTorchModel,
    SingleTaskVariationalGP,
)
from botorch.models.cost import AffineFidelityCostModel
from botorch.models.deterministic import (
    AffineDeterministicModel,
    GenericDeterministicModel,
    PosteriorMeanModel,
)
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.fully_bayesian_multitask import SaasFullyBayesianMultiTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gp_regression_fidelity import SingleTaskMultiFidelityGP
from botorch.models.gp_regression_mixed import MixedSingleTaskGP
from botorch.models.higher_order_gp import HigherOrderGP
from botorch.models.map_saas import (
    add_saas_prior,
    AdditiveMapSaasSingleTaskGP,
    EnsembleMapSaasSingleTaskGP,
)
from botorch.models.model import ModelList
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.multitask import KroneckerMultiTaskGP, MultiTaskGP
from botorch.models.pairwise_gp import PairwiseGP, PairwiseLaplaceMarginalLogLikelihood

__all__ = [
    "add_saas_prior",
    "AdditiveMapSaasSingleTaskGP",
    "AffineDeterministicModel",
    "AffineFidelityCostModel",
    "ApproximateGPyTorchModel",
    "EnsembleMapSaasSingleTaskGP",
    "GenericDeterministicModel",
    "HigherOrderGP",
    "KroneckerMultiTaskGP",
    "MixedSingleTaskGP",
    "ModelList",
    "ModelListGP",
    "MultiTaskGP",
    "PairwiseGP",
    "PairwiseLaplaceMarginalLogLikelihood",
    "PosteriorMeanModel",
    "SaasFullyBayesianMultiTaskGP",
    "SaasFullyBayesianSingleTaskGP",
    "SingleTaskGP",
    "SingleTaskMultiFidelityGP",
    "SingleTaskVariationalGP",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Multi-objective multi-fidelity optimization benchmark problems.

References

.. [Irshad2021]
    F. Irshad, S. Karsch, and A. Döpp. Expected hypervolume improvement for
    simultaneous multi-objective and multi-fidelity optimization.
    arXiv preprint arXiv:2112.13901, 2021.
"""

import math

import torch
from botorch.test_functions.base import MultiObjectiveTestProblem
from torch import Tensor


class MOMFBraninCurrin(MultiObjectiveTestProblem):
    r"""Branin-Currin problem for multi-objective-multi-fidelity optimization.

    (2+1)-dimensional function with domain `[0,1]^3` where the last dimension
    is the fidelity parameter `s`.
    Both functions assume minimization. See [Irshad2021]_ for more details.

    Modified Branin function:

        B(x,s) = 21-((
        15*x_2 - b(s) * (15 * x_1 - 5) ** 2 + c(s) * (15 * x_1 - 5) - 6 ) ** 2
        + 10 * (1 - t(s)) * cos(15 * x_1 - 5)+10)/22

    Here `b`, `c`, `r` and `t` are constants and `s` is the fidelity parameter:
        where `b = 5.1 / (4 * math.pi ** 2) - 0.01(1-s)`,
        `c = 5 / math.pi - 0.1*(1 - s)`,
        `r = 6`,
        `t = 1 / (8 * math.pi) + 0.05*(1-s)`

    Modified Currin function:

        C(x) = 14-((1 - 0.1(1-s)exp(-1 / (2 * x_2))) * (
        2300 * x_1 ** 3 + 1900 * x_1 ** 2 + 2092 * x_1 + 60
        ) / 100 * x_1 ** 3 + 500 * x_1 ** 2 + 4 * x_2 + 20)/15

    """

    dim = 3
    continuous_inds = list(range(3))
    num_objectives = 2
    _bounds = [(0.0, 1.0) for _ in range(dim)]
    _ref_point = [0, 0]
    _max_hv = 0.5235514158034145

    def _branin(self, X: Tensor) -> Tensor:
        x1 = X[..., 0]
        x2 = X[..., 1]
        s = X[..., 2]

        x11 = 15 * x1 - 5
        x22 = 15 * x2
        b = 5.1 / (4 * math.pi**2) - 0.01 * (1 - s)
        c = 5 / math.pi - 0.1 * (1 - s)
        r = 6
        t = 1 / (8 * math.pi) + 0.05 * (1 - s)
        y = (x22 - b * x11**2 + c * x11 - r) ** 2 + 10 * (1 - t) * torch.cos(x11) + 10
        B = 21 - y
        return B / 22

    def _currin(self, X: Tensor) -> Tensor:
        x1 = X[..., 0]
        x2 = X[..., 1]
        s = X[..., 2]
        A = 2300 * x1**3 + 1900 * x1**2 + 2092 * x1 + 60
        B = 100 * x1**3 + 500 * x1**2 + 4 * x1 + 20
        y = (1 - 0.1 * (1 - s) * torch.exp(-1 / (2 * x2))) * A / B
        C = -y + 14
        return C / 15

    def _evaluate_true(self, X: Tensor) -> Tensor:
        branin = self._branin(X)
        currin = self._currin(X)
        return torch.stack([-branin, -currin], dim=-1)


class MOMFPark(MultiObjectiveTestProblem):
    r"""Modified Park test functions for multi-objective multi-fidelity optimization.

    (4+1)-dimensional function with domain `[0,1]^5` where the last dimension
    is the fidelity parameter `s`. See [Irshad2021]_ for more details.

    The first modified Park function is

        P1(x, s)=A*(T1(x,s)+T2(x,s)-B)/22-0.8

    The second modified Park function is

        P2(x,s)=A*(5-2/3*exp(x1+x2)-x4*sin(x3)*A+x3-B)/4 - 0.7

    Here

        T_1(x,s) = (x1+0.001*(1-s))/2*sqrt(1+(x2+x3**2)*x4/(x1**2))

        T_2(x, s) = (x1+3*x4)*exp(1+sin(x3))

    and `A(s)=(0.9+0.1*s)`, `B(s)=0.1*(1-s)`.
    """

    dim = 5
    continuous_inds = list(range(5))
    num_objectives = 2
    _bounds = [(0.0, 1.0) for _ in range(dim)]
    _ref_point = [0, 0]
    _max_hv = 0.08551927363087991

    def _transform(self, X: Tensor) -> Tensor:
        x1 = X[..., 0]
        x2 = X[..., 1]
        x3 = X[..., 2]
        x4 = X[..., 3]
        s = X[..., 4]
        _x1 = 1 - 2 * (x1 - 0.6) ** 2
        _x2 = x2
        _x3 = 1 - 3 * (x3 - 0.5) ** 2
        _x4 = 1 - (x4 - 0.8) ** 2
        return torch.stack([_x1, _x2, _x3, _x4, s], dim=-1)

    def _park1(self, X: Tensor) -> Tensor:
        x1 = X[..., 0]
        x2 = X[..., 1]
        x3 = X[..., 2]
        x4 = X[..., 3]
        s = X[..., 4]
        T1 = (
            (x1 + 1e-3 * (1 - s))
            / 2
            * torch.sqrt(1 + (x2 + x3**2) * x4 / (x1**2 + 1e-4))
        )
        T2 = (x1 + 3 * x4) * torch.exp(1 + torch.sin(x3))
        A = 0.9 + 0.1 * s
        B = 0.1 * (1 - s)
        return A * (T1 + T2 - B) / 22 - 0.8

    def _park2(self, X: Tensor) -> Tensor:
        x1 = X[..., 0]
        x2 = X[..., 1]
        x3 = X[..., 2]
        x4 = X[..., 3]
        s = X[..., 4]
        A = 0.9 + 0.1 * s
        B = 0.1 * (1 - s)
        return (
            A * (5 - 2 / 3 * torch.exp(x1 + x2) + x4 * torch.sin(x3) * A - x3 + B) / 4
            - 0.7
        )

    def _evaluate_true(self, X: Tensor) -> Tensor:
        X = self._transform(X)
        park1 = self._park1(X)
        park2 = self._park2(X)
        return torch.stack([-park1, -park2], dim=-1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from abc import ABC, abstractmethod

import torch
import torch.nn as nn
from botorch.models.model import Model
from botorch_community.models.vbll_helper import DenseNormal, Normal
from torch import Tensor


class AbstractBLLModel(Model, ABC):
    def __init__(self):
        """Abstract class for Bayesian Last Layer (BLL) models."""
        super().__init__()
        self.model = None

    @property
    def num_outputs(self) -> int:
        return self.model.num_outputs

    @property
    def num_inputs(self):
        return self.model.num_inputs

    @property
    def device(self):
        return self.model.device

    @abstractmethod
    def __call__(self, X: Tensor) -> Normal | DenseNormal:
        pass  # pragma: no cover

    @abstractmethod
    def fit(self, *args, **kwargs):
        pass  # pragma: no cover

    @abstractmethod
    def sample(self, sample_shape: torch.Size | None = None) -> nn.Module:
        pass  # pragma: no cover



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from itertools import product

import torch
from botorch.utils.test_helpers import get_fully_bayesian_model
from botorch.utils.testing import BotorchTestCase
from botorch_community.acquisition.bayesian_active_learning import (
    qBayesianQueryByComittee,
    qBayesianVarianceReduction,
    qStatisticalDistanceActiveLearning,
)


class TestQStatisticalDistanceActiveLearning(BotorchTestCase):
    def test_q_statistical_distance_active_learning(self):
        torch.manual_seed(1)
        tkwargs = {"device": self.device}
        distance_metrics = ("hellinger", "kl_divergence")
        num_objectives = 1
        num_models = 3
        for (
            dtype,
            distance_metric,
            standardize_model,
            infer_noise,
        ) in product(
            (torch.float, torch.double),
            distance_metrics,
            (False, True),  # standardize_model
            (True,),  # infer_noise - only one option avail in PyroModels
        ):
            tkwargs["dtype"] = dtype
            input_dim = 2
            train_X = torch.rand(4, input_dim, **tkwargs)
            train_Y = torch.rand(4, num_objectives, **tkwargs)

            model = get_fully_bayesian_model(
                train_X=train_X,
                train_Y=train_Y,
                num_models=num_models,
                standardize_model=standardize_model,
                infer_noise=infer_noise,
                **tkwargs,
            )

            # test acquisition
            X_pending_list = [None, torch.rand(2, input_dim, **tkwargs)]
            for i in range(len(X_pending_list)):
                X_pending = X_pending_list[i]

                acq = qStatisticalDistanceActiveLearning(
                    model=model,
                    X_pending=X_pending,
                    distance_metric=distance_metric,
                )

                test_Xs = [
                    torch.rand(4, 1, input_dim, **tkwargs),
                    torch.rand(4, 3, input_dim, **tkwargs),
                    torch.rand(4, 5, 1, input_dim, **tkwargs),
                    torch.rand(4, 5, 3, input_dim, **tkwargs),
                ]

                for j in range(len(test_Xs)):
                    acq_X = acq.forward(test_Xs[j])
                    acq_X = acq(test_Xs[j])
                    # assess shape
                    self.assertTrue(acq_X.shape == test_Xs[j].shape[:-2])

        with self.assertRaises(ValueError):
            acq = qStatisticalDistanceActiveLearning(
                model=model,
                distance_metric="NOT_A_DISTANCE",
                X_pending=X_pending,
            )


class TestQBayesianQueryByComittee(BotorchTestCase):
    def test_q_bayesian_query_by_comittee(self):
        torch.manual_seed(1)
        tkwargs = {"device": self.device}
        num_objectives = 1
        num_models = 3
        for (
            dtype,
            standardize_model,
            infer_noise,
        ) in product(
            (torch.float, torch.double),
            (False, True),  # standardize_model
            (True,),  # infer_noise - only one option avail in PyroModels
        ):
            tkwargs["dtype"] = dtype
            input_dim = 2
            train_X = torch.rand(4, input_dim, **tkwargs)
            train_Y = torch.rand(4, num_objectives, **tkwargs)

            model = get_fully_bayesian_model(
                train_X=train_X,
                train_Y=train_Y,
                num_models=num_models,
                standardize_model=standardize_model,
                infer_noise=infer_noise,
                **tkwargs,
            )

            # test acquisition
            X_pending_list = [None, torch.rand(2, input_dim, **tkwargs)]
            for i in range(len(X_pending_list)):
                X_pending = X_pending_list[i]

                acq = qBayesianQueryByComittee(
                    model=model,
                    X_pending=X_pending,
                )

                test_Xs = [
                    torch.rand(4, 1, input_dim, **tkwargs),
                    torch.rand(4, 3, input_dim, **tkwargs),
                    torch.rand(4, 5, 1, input_dim, **tkwargs),
                    torch.rand(4, 5, 3, input_dim, **tkwargs),
                ]

                for j in range(len(test_Xs)):
                    acq_X = acq.forward(test_Xs[j])
                    acq_X = acq(test_Xs[j])
                    # assess shape
                    self.assertTrue(acq_X.shape == test_Xs[j].shape[:-2])


class TestQBayesianVarianceReduction(BotorchTestCase):
    def test_q_bayesian_variance_reduction(self):
        torch.manual_seed(1)
        tkwargs = {"device": self.device}
        num_objectives = 1
        num_models = 3
        for (
            dtype,
            standardize_model,
            infer_noise,
        ) in product(
            (torch.float, torch.double),
            (False, True),  # standardize_model
            (True,),  # infer_noise - only one option avail in PyroModels
        ):
            tkwargs["dtype"] = dtype
            input_dim = 2
            train_X = torch.rand(4, input_dim, **tkwargs)
            train_Y = torch.rand(4, num_objectives, **tkwargs)

            model = get_fully_bayesian_model(
                train_X=train_X,
                train_Y=train_Y,
                num_models=num_models,
                standardize_model=standardize_model,
                infer_noise=infer_noise,
                **tkwargs,
            )

            # test acquisition
            X_pending_list = [None, torch.rand(2, input_dim, **tkwargs)]
            for i in range(len(X_pending_list)):
                X_pending = X_pending_list[i]

                acq = qBayesianVarianceReduction(
                    model=model,
                    X_pending=X_pending,
                )

                test_Xs = [
                    torch.rand(4, 1, input_dim, **tkwargs),
                    torch.rand(4, 3, input_dim, **tkwargs),
                    torch.rand(4, 5, 1, input_dim, **tkwargs),
                    torch.rand(4, 5, 3, input_dim, **tkwargs),
                ]

                for j in range(len(test_Xs)):
                    acq_X = acq.forward(test_Xs[j])
                    acq_X = acq(test_Xs[j])
                    # assess shape
                    self.assertTrue(acq_X.shape == test_Xs[j].shape[:-2])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Base class for test functions for optimization benchmarks.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterable, Iterator, Protocol

import torch
from botorch.exceptions.errors import InputDataError, UnsupportedError
from pyre_extensions import none_throws
from torch import Tensor
from torch.nn import Module


def validate_parameter_indices(
    dim: int,
    bounds: Tensor,
    continuous_inds: list[int],
    discrete_inds: list[int],
    categorical_inds: list[int],
) -> None:
    r"""Check that the parameter indices are valid.

    Args:
        dim: Number of search space dimensions.
        bounds: A `2 x d`-dim tensor of lower and upper bounds.
        continuous_inds: List of unique integers corresponding to continuous parameters.
        discrete_inds: List of unique integers corresponding to discrete parameters.
        categorical_inds: List of unique integers corresponding to categorical
            parameters.

    Raises:
        ValueError: If the parameter indices are invalid.
    """
    for inds in [continuous_inds, discrete_inds, categorical_inds]:
        if len(inds) == 0:
            continue  # Nothing to check
        if not (
            isinstance(inds, list)
            and all(isinstance(x, int) for x in inds)
            and len(set(inds)) == len(inds)
            and min(inds) >= 0
            and max(inds) <= dim - 1
        ):
            raise ValueError(
                "All parameter indices must be a list with unique integers between "
                f"0 and dim - 1. Got {inds=}."
            )
    # Let's make sure all parameters are covered.
    all_inds = continuous_inds + discrete_inds + categorical_inds
    if (len(all_inds) != dim) or (set(all_inds) != set(range(dim))):
        raise ValueError(
            f"All parameter indices must be present, got {dim=}, {continuous_inds=}, "
            f"{discrete_inds=}, {categorical_inds=}"
        )
    # Check the shape of the bounds
    if not bounds.shape == (2, dim):
        raise ValueError(
            f"Expected `bounds` to have shape `2 x d`. Got {bounds.shape=}, {dim=}."
        )
    # Check that the bounds are integer valued for discrete and categorical parameters.
    for inds in [discrete_inds, categorical_inds]:
        if len(inds) > 0 and (not (bounds[:, inds] == bounds[:, inds].round()).all()):
            raise ValueError(
                "Expected the lower and upper bounds of the discrete and categorical "
                "parameters to be integer-valued."
            )


def validate_inputs(
    X: Tensor,
    dim: int,
    bounds: Tensor,
    discrete_inds: list[int],
    categorical_inds: list[int],
) -> None:
    r"""Check that the inputs are valid.

    This method checks that the input tensor `X` has the correct shape, is within
    the bounds, and that the discrete and categorical parameters are integer-valued.

    Args:
        X: A `(batch_shape) x n x d`-dim tensor of point(s) at which to evaluate
        dim: Number of search space dimensions.
        bounds: A `2 x d`-dim tensor of lower and upper bounds.
        discrete_inds: List of unique integers corresponding to discrete parameters.
        categorical_inds: List of unique integers corresponding to categorical
            parameters.

    Raises:
        ValueError: If the parameter indices are invalid.
    """

    if not X.shape[-1] == dim:
        raise ValueError(
            "Expected `X` to have shape `(batch_shape) x n x d`. "
            f"Got {X.shape=} and {dim=}"
        )
    if not ((X >= bounds[0]).all() and (X <= bounds[1]).all()):
        raise ValueError("Expected `X` to be within the bounds of the test problem.")
    for inds in [discrete_inds, categorical_inds]:
        if not (X[..., inds] == X[..., inds].round()).all():
            raise ValueError(
                "Expected `X` to have integer values for the discrete and "
                "categorical parameters."
            )


class BaseTestProblem(Module, ABC):
    r"""Base class for test functions."""

    dim: int
    _bounds: list[
        tuple[float, float]
    ]  # Bounds, must be integers for discrete/categorical parameters
    _check_grad_at_opt: bool = True
    continuous_inds: list[int] = []  # Float-valued range parameters (bounds inclusive)
    discrete_inds: list[int] = []  # Ordered integer parameters (bounds inclusive)
    categorical_inds: list[int] = []  # Unordered integer parameters (bounds inclusive)
    # Whether the problem is a minimization problem by default, with `negate=False`.
    _is_minimization_by_default: bool = True

    def __init__(
        self,
        noise_std: None | float | list[float] = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""Base constructor for test functions.

        Args:
            noise_std: Standard deviation of the observation noise. If a list is
                provided, specifies separate noise standard deviations for each
                objective in a multiobjective problem.
            negate: If True, negate the function.
            dtype: The dtype that is used for the bounds of the function.
        """
        super().__init__()
        self.noise_std = noise_std
        self.negate = negate
        if len(self._bounds) != self.dim:
            raise InputDataError(
                "Expected the bounds to match the dimensionality of the domain. "
                f"Got {self.dim=} and {len(self._bounds)=}."
            )
        self.register_buffer(
            "bounds",
            torch.tensor(self._bounds, dtype=dtype).transpose(-1, -2),
        )
        validate_parameter_indices(
            dim=self.dim,
            bounds=self.bounds,
            continuous_inds=self.continuous_inds,
            discrete_inds=self.discrete_inds,
            categorical_inds=self.categorical_inds,
        )

    def forward(self, X: Tensor, noise: bool = True) -> Tensor:
        r"""Evaluate the function on a set of points.

        Args:
            X: A `(batch_shape) x d`-dim tensor of point(s) at which to evaluate
                the function.
            noise: If `True`, add observation noise as specified by `noise_std`.

        Returns:
            A `batch_shape`-dim tensor ouf function evaluations.
        """
        f = self.evaluate_true(X=X)
        if noise and self.noise_std is not None:
            _noise = torch.tensor(self.noise_std, device=X.device, dtype=X.dtype)
            f += _noise * torch.randn_like(f)
        if self.negate:
            f = -f
        return f

    def evaluate_true(self, X: Tensor) -> Tensor:
        r"""
        Evaluate the function (w/o observation noise) on a set of points.

        Args:
            X: A `(batch_shape) x d`-dim tensor of point(s) at which to
                evaluate.

        Returns:
            A `batch_shape`-dim tensor.
        """
        validate_inputs(
            X=X,
            dim=self.dim,
            bounds=self.bounds,
            discrete_inds=self.discrete_inds,
            categorical_inds=self.categorical_inds,
        )
        return self._evaluate_true(X=X)

    @abstractmethod
    def _evaluate_true(self, X: Tensor) -> Tensor:
        r"""Evaluate the function (w/o observation noise) on a set of points.

        Args:
            X: A `(batch_shape) x d`-dim tensor of point(s) at which to
                evaluate.

        Returns:
            A `batch_shape`-dim tensor.
        """
        pass  # pragma: no cover

    @property
    def is_minimization_problem(self) -> bool:
        r"""Whether the problem is a minimization problem, after accounting
        for the `negate` option.
        """
        return (
            self._is_minimization_by_default
            if not self.negate
            else not self._is_minimization_by_default
        )


class ConstrainedBaseTestProblem(BaseTestProblem, ABC):
    r"""Base class for test functions with constraints.

    In addition to one or more objectives, a problem may have a number of outcome
    constraints of the form `c_i(x) >= 0` for `i=1, ..., n_c`.

    This base class provides common functionality for such problems.
    """

    num_constraints: int
    _check_grad_at_opt: bool = False
    constraint_noise_std: None | float | list[float] = None
    _worst_feasible_value: float | None = None

    def evaluate_slack(self, X: Tensor, noise: bool = True) -> Tensor:
        r"""Evaluate the constraint slack on a set of points.

        Constraints `i` is assumed to be feasible at `x` if the associated slack
        `c_i(x)` is positive. Zero slack means that the constraint is active. Negative
        slack means that the constraint is violated.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraint slacks: `c_1(X), ...., c_{n_c}(X)`.
            noise: If `True`, add observation noise to the slack as specified by
                `noise_std`.

        Returns:
            A `batch_shape x n_c`-dim tensor of constraint slack (where positive slack
                corresponds to the constraint being feasible).
        """
        cons = self.evaluate_slack_true(X=X)
        if noise and self.constraint_noise_std is not None:
            _constraint_noise = torch.tensor(
                self.constraint_noise_std, device=X.device, dtype=X.dtype
            )
            cons += _constraint_noise * torch.randn_like(cons)
        return cons

    def is_feasible(self, X: Tensor, noise: bool = True) -> Tensor:
        r"""Evaluate whether the constraints are feasible on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraints.
            noise: If `True`, add observation noise as specified by `noise_std`.

        Returns:
            A `batch_shape`-dim boolean tensor that is `True` iff all constraint
                slacks (potentially including observation noise) are positive.
        """
        return (self.evaluate_slack(X=X, noise=noise) >= 0.0).all(dim=-1)

    def evaluate_slack_true(self, X: Tensor) -> Tensor:
        r"""Evaluate the constraint slack (w/o observation noise) on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraint slacks: `c_1(X), ...., c_{n_c}(X)`.

        Returns:
            A `batch_shape x n_c`-dim tensor of constraint slack (where positive slack
                corresponds to the constraint being feasible).
        """
        validate_inputs(
            X=X,
            dim=self.dim,
            bounds=self.bounds,
            discrete_inds=self.discrete_inds,
            categorical_inds=self.categorical_inds,
        )
        return self._evaluate_slack_true(X=X)

    @abstractmethod
    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        r"""Evaluate the constraint slack (w/o observation noise) on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraint slacks: `c_1(X), ...., c_{n_c}(X)`.

        Returns:
            A `batch_shape x n_c`-dim tensor of constraint slack (where positive slack
                corresponds
        """
        pass  # pragma: no cover

    @property
    def worst_feasible_value(self) -> float:
        r"""The worst feasible value of the objective function. This is useful when
        evaluating the performance of different optimization methods as this value
        can be assigned to all infeasible trials. This has the desirable property that
        any feasible trial has better performance than an infeasible trial.
        """
        if isinstance(self, MultiObjectiveTestProblem):
            return 0.0  # Can return 0.0 for MOO since this is the smallest possible HV
        elif self._worst_feasible_value is not None:
            return (
                -self._worst_feasible_value
                if self.negate
                else self._worst_feasible_value
            )
        raise NotImplementedError(
            f"Problem {self.__class__.__name__} does not specify the "
            "worst feasible value."
        )


class MultiObjectiveTestProblem(BaseTestProblem, ABC):
    r"""Base class for multi-objective test functions.

    TODO: add a pareto distance function that returns the distance
    between a provided point and the closest point on the true pareto front.
    """

    num_objectives: int
    _ref_point: list[float]
    _max_hv: float | None = None

    def __init__(
        self,
        noise_std: None | float | list[float] = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""Base constructor for multi-objective test functions.

        Args:
            noise_std: Standard deviation of the observation noise. If a list is
                provided, specifies separate noise standard deviations for each
                objective.
            negate: If True, negate the objectives.
        """
        if isinstance(noise_std, list) and len(noise_std) != len(self._ref_point):
            raise InputDataError(
                f"If specified as a list, length of noise_std ({len(noise_std)}) "
                f"must match the number of objectives ({len(self._ref_point)})"
            )
        super().__init__(noise_std=noise_std, negate=negate, dtype=dtype)
        ref_point = torch.tensor(self._ref_point, dtype=dtype)
        if negate:
            ref_point *= -1
        self.register_buffer("ref_point", ref_point)

    @property
    def max_hv(self) -> float:
        if self._max_hv is not None:
            return self._max_hv
        else:
            raise NotImplementedError(
                f"Problem {self.__class__.__name__} does not specify maximal "
                "hypervolume."
            )

    def gen_pareto_front(self, n: int) -> Tensor:
        r"""Generate `n` pareto optimal points."""
        raise NotImplementedError

    @property
    def is_minimization_problem(self) -> bool:
        raise UnsupportedError(
            "`is_minimization_problem` is not a valid property for "
            "multi-objective problems. "
        )


class SeedingMixin(ABC):
    _seeds: Iterator[int] | None
    _current_seed: int | None

    @property
    def has_seeds(self) -> bool:
        return self._seeds is not None

    def increment_seed(self) -> int:
        self._current_seed = next(none_throws(self._seeds))
        return none_throws(self._current_seed)

    @property
    def seed(self) -> int | None:
        return self._current_seed


# Outlier problems
class OutlierGenerator(Protocol):
    def __call__(self, problem: BaseTestProblem, X: Tensor, bounds: Tensor) -> Tensor:
        """Call signature for outlier generators for single-objective problems.

        Args:
            problem: The test problem.
            X: The input tensor.
            bounds: The bounds of the test problem.

        Returns:
            A tensor of outliers with shape X.shape[:-1] (1d if unbatched).
        """
        pass  # pragma: no cover


def constant_outlier_generator(
    problem: Any, X: Tensor, bounds: Any, constant: float
) -> Tensor:
    """
    Generates outliers that are all the same constant. To be used in conjunction with
    `partial` to fix the constant value and conform to the `OutlierGenerator` protocol.

    Example:
        >>> generator = partial(constant_outlier_generator, constant=1.0)

    Args:
        problem: Not used.
        X: The `batch_shape x n x d`-dim inputs. Also determines the number, dtype,
            and device of the returned tensor.
        bounds: Not used.
        constant: The constant value of the outliers.

    Returns:
        Tensor of shape `batch_shape x n` (1d if unbatched).
    """
    return torch.full(X.shape[:-1], constant, dtype=X.dtype, device=X.device)


class CorruptedTestProblem(BaseTestProblem, SeedingMixin):
    def __init__(
        self,
        base_test_problem: BaseTestProblem,
        outlier_generator: OutlierGenerator,
        outlier_fraction: float,
        bounds: list[tuple[float, float]] | None = None,
        seeds: Iterable[int] | None = None,
    ) -> None:
        """A problem with outliers.

        NOTE: Both noise_std and negate will be taken from the base test problem.

        Args:
            base_test_problem: The base function to be corrupted.
            outlier_generator: A function that generates outliers. It will be called
                with arguments `f`, `X` and `bounds`, where `f` is the
                `base_test_problem`, `X` is the
                argument passed to the `forward` method, and `bounds`
                are as here, and it returns the values of outliers.
            outlier_fraction: The fraction of outliers.
            bounds: The bounds of the function.
            seeds: The seeds to use for the outlier generator. If seeds are provided,
                the problem will iterate through the list of seeds, changing the seed
                with a call to `next(seeds)` with every `forward` call. If a list is
                provided, it will first be converted to an iterator.
        """
        self.dim: int = base_test_problem.dim
        self.continuous_inds = base_test_problem.continuous_inds
        self.discrete_inds = base_test_problem.discrete_inds
        self.categorical_inds = base_test_problem.categorical_inds
        self._bounds: list[tuple[float, float]] = (
            bounds if bounds is not None else base_test_problem._bounds
        )
        super().__init__(
            noise_std=base_test_problem.noise_std,
            negate=base_test_problem.negate,
        )
        self.base_test_problem = base_test_problem
        self.outlier_generator = outlier_generator
        self.outlier_fraction = outlier_fraction
        self._current_seed: int | None = None
        self._seeds: Iterator[int] | None = None if seeds is None else iter(seeds)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return self.base_test_problem.evaluate_true(X)

    def forward(self, X: Tensor, noise: bool = True) -> Tensor:
        """
        Generate data at X and corrupt it, if noise is True.

        Args:
            X: The `batch_shape x n x d`-dim inputs.
            noise: Whether to corrupt the data.

        Returns:
            A `batch_shape x n`-dim tensor.
        """
        Y = super().forward(X, noise=noise)
        if noise:
            if self.has_seeds:
                self.increment_seed()
                torch.manual_seed(self.seed)
            corrupt = torch.rand(X.shape[:-1]) < self.outlier_fraction
            outliers = self.outlier_generator(
                problem=self.base_test_problem, X=X, bounds=self.bounds
            )
            Y = torch.where(corrupt, outliers, Y)
        return Y



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from abc import abstractmethod

import torch
from botorch.acquisition.objective import GenericMCObjective, MCAcquisitionObjective
from botorch.exceptions.errors import BotorchError, BotorchTensorDimensionError
from botorch.models.model import Model
from botorch.utils import apply_constraints
from botorch.utils.transforms import normalize_indices
from torch import Tensor


class MCMultiOutputObjective(MCAcquisitionObjective):
    r"""Abstract base class for MC multi-output objectives.

    Args:
        _is_mo: A boolean denoting whether the objectives are multi-output.
    """

    _is_mo: bool = True

    @abstractmethod
    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Evaluate the multi-output objective on the samples.

        Args:
            samples: A `sample_shape x batch_shape x q x m`-dim Tensors of samples from
                a model posterior.
            X: A `batch_shape x q x d`-dim Tensors of inputs.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim Tensor of objective values with
            `m'` the output dimension. This assumes maximization in each output
            dimension).

        This method is usually not called directly, but via the objectives.

        Example:
            >>> # `__call__` method:
            >>> samples = sampler(posterior)
            >>> outcomes = multi_obj(samples)
        """
        pass  # pragma: no cover


class GenericMCMultiOutputObjective(GenericMCObjective, MCMultiOutputObjective):
    r"""Multi-output objective generated from a generic callable.

    Allows to construct arbitrary MC-objective functions from a generic
    callable. In order to be able to use gradient-based acquisition function
    optimization it should be possible to backpropagate through the callable.
    """

    pass


class IdentityMCMultiOutputObjective(MCMultiOutputObjective):
    r"""Trivial objective that returns the unaltered samples.

    Example:
        >>> identity_objective = IdentityMCMultiOutputObjective()
        >>> samples = sampler(posterior)
        >>> objective = identity_objective(samples)
    """

    def __init__(
        self, outcomes: list[int] | None = None, num_outcomes: int | None = None
    ) -> None:
        r"""Initialize Objective.

        Args:
            outcomes: A list of the `m'` indices that the weights should be
                applied to.
            num_outcomes: The total number of outcomes `m`
        """
        super().__init__()
        if outcomes is not None:
            if len(outcomes) < 2:
                raise BotorchTensorDimensionError(
                    "Must specify at least two outcomes for MOO."
                )
            if any(i < 0 for i in outcomes):
                if num_outcomes is None:
                    raise BotorchError(
                        "num_outcomes is required if any outcomes are less than 0."
                    )
                outcomes = normalize_indices(outcomes, num_outcomes)
            self.register_buffer("outcomes", torch.tensor(outcomes, dtype=torch.long))

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        if hasattr(self, "outcomes"):
            return samples.index_select(-1, self.outcomes.to(device=samples.device))
        return samples


class WeightedMCMultiOutputObjective(IdentityMCMultiOutputObjective):
    r"""Objective that reweights samples by given weights vector.

    Example:
        >>> weights = torch.tensor([1.0, -1.0])
        >>> weighted_objective = WeightedMCMultiOutputObjective(weights)
        >>> samples = sampler(posterior)
        >>> objective = weighted_objective(samples)
    """

    def __init__(
        self,
        weights: Tensor,
        outcomes: list[int] | None = None,
        num_outcomes: int | None = None,
    ) -> None:
        r"""Initialize Objective.

        Args:
            weights: `m'`-dim tensor of outcome weights.
            outcomes: A list of the `m'` indices that the weights should be
                applied to.
            num_outcomes: the total number of outcomes `m`
        """
        super().__init__(outcomes=outcomes, num_outcomes=num_outcomes)
        if weights.ndim != 1:
            raise BotorchTensorDimensionError(
                f"weights must be an 1-D tensor, but got {weights.shape}."
            )
        elif outcomes is not None and weights.shape[0] != len(outcomes):
            raise BotorchTensorDimensionError(
                "weights must contain the same number of elements as outcomes, "
                f"but got {weights.numel()} weights and {len(outcomes)} outcomes."
            )
        self.register_buffer("weights", weights)

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        samples = super().forward(samples=samples)
        return samples * self.weights.to(samples)


class FeasibilityWeightedMCMultiOutputObjective(MCMultiOutputObjective):
    def __init__(
        self,
        model: Model,
        X_baseline: Tensor,
        constraint_idcs: list[int],
        objective: MCMultiOutputObjective | None = None,
    ) -> None:
        r"""Construct a feasibility-weighted objective.

        This applies feasibility weighting before calculating the objective value.
        Defaults to identity if no constraints or objective is present.

        NOTE: By passing in a single-output `MCAcquisitionObjective` as the `objective`,
        this can be used as a single-output `MCAcquisitionObjective` as well.

        Args:
            model: A fitted Model.
            X_baseline: An `n x d`-dim tensor of points already observed.
            constraint_idcs: The outcome indices of the constraints. Constraints are
                handled by weighting the samples according to a sigmoid approximation
                of feasibility. A positive constraint outcome implies feasibility.
            objective: An optional objective to apply after feasibility-weighting
                the samples.
        """
        super().__init__()
        num_outputs = model.num_outputs
        # Get the non-negative indices.
        constraint_idcs = [
            num_outputs + idx if idx < 0 else idx for idx in constraint_idcs
        ]
        if len(constraint_idcs) != len(set(constraint_idcs)):
            raise ValueError("Received duplicate entries for `constraint_idcs`.")
        # Extract the indices for objective outcomes.
        objective_idcs = [i for i in range(num_outputs) if i not in constraint_idcs]
        if len(constraint_idcs) > 0:
            # Import locally to avoid circular import.
            from botorch.acquisition.utils import get_infeasible_cost

            inf_cost = get_infeasible_cost(
                X=X_baseline, model=model, objective=lambda y, X: y
            )[objective_idcs]

            def apply_feasibility_weights(Y: Tensor, X: Tensor | None = None) -> Tensor:
                return apply_constraints(
                    obj=Y[..., objective_idcs],
                    constraints=[lambda Y: -Y[..., i] for i in constraint_idcs],
                    samples=Y,
                    # This ensures that the dtype/device is set properly.
                    infeasible_cost=inf_cost.to(Y),
                )

            self.apply_feasibility_weights = apply_feasibility_weights
        else:
            self.apply_feasibility_weights = lambda Y: Y
        if objective is None:
            self.objective = lambda Y, X: Y
        else:
            self.objective = objective
            self._verify_output_shape = objective._verify_output_shape

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        return self.objective(self.apply_feasibility_weights(samples), X=X)
